import logging

from django.conf import settings
from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _
from django.views.generic import (
    FormView as DjangoFormView, DetailView, TemplateView
)
from django.views.generic.base import View
from django.views.generic.detail import SingleObjectMixin
from django.views.generic.edit import (
    CreateView, DeleteView, FormMixin, ModelFormMixin, UpdateView
)
from django.views.generic.list import ListView

from pure_pagination.mixins import PaginationMixin

from mayan.apps.acls.models import AccessControlList

from .forms import ChoiceForm
from .icons import (
    icon_add_all, icon_confirm_form_cancel, icon_confirm_form_submit,
    icon_remove_all, icon_assign_remove_add, icon_assign_remove_remove
)
from .mixins import (
    ExtraDataDeleteViewMixin, DownloadViewMixin, DynamicFormViewMixin,
    ExternalObjectViewMixin, ExtraContextViewMixin, FormExtraKwargsViewMixin,
    ListModeViewMixin, MultipleObjectViewMixin, ObjectActionViewMixin,
    ObjectNameViewMixin, RedirectionViewMixin, RestrictedQuerysetViewMixin,
    SortingViewMixin, ViewPermissionCheckViewMixin
)

from .settings import setting_paginate_by

logger = logging.getLogger(name=__name__)


# Required by other views, moved to the top
class MultiFormView(DjangoFormView):
    form_extra_kwargs = None
    prefix = None
    prefixes = {}
    template_name = 'appearance/generic_form.html'

    def _create_form(self, form_name, klass):
        form_kwargs = self.get_form_kwargs(form_name=form_name)
        form_create_method = 'create_{}_form'.format(form_name)
        if hasattr(self, form_create_method):
            form = getattr(self, form_create_method)(**form_kwargs)
        else:
            form = klass(**form_kwargs)
        return form

    def all_forms_valid(self, forms):
        return None

    def dispatch(self, request, *args, **kwargs):
        form_classes = self.get_form_classes()
        self.forms = self.get_forms(form_classes=form_classes)
        return super().dispatch(request, *args, **kwargs)

    def forms_invalid(self, forms):
        return self.render_to_response(self.get_context_data(forms=forms))

    def forms_valid(self, forms):
        for form_name, form in forms.items():
            form_valid_method = '{}_form_valid'.format(form_name)

            if hasattr(self, form_valid_method):
                return getattr(self, form_valid_method)(form=form)

        self.all_forms_valid(forms=forms)

        return HttpResponseRedirect(redirect_to=self.get_success_url())

    def get_context_data(self, **kwargs):
        """
        Insert the form into the context dict.
        """
        if 'forms' not in kwargs:
            kwargs['forms'] = self.get_forms(
                form_classes=self.get_form_classes()
            )
        return super(FormMixin, self).get_context_data(**kwargs)

    def get_form_classes(self):
        return self.form_classes

    def get_form_extra_kwargs(self, form_name):
        return self.form_extra_kwargs

    def get_form_kwargs(self, form_name):
        kwargs = {}
        kwargs.update({'initial': self.get_initial(form_name=form_name)})
        kwargs.update({'prefix': self.get_prefix(form_name=form_name)})

        if self.request.method in ('POST', 'PUT'):
            kwargs.update(
                {
                    'data': self.request.POST,
                    'files': self.request.FILES,
                }
            )

        kwargs.update(self.get_form_extra_kwargs(form_name=form_name) or {})

        return kwargs

    def get_forms(self, form_classes):
        return dict(
            [
                (
                    key, self._create_form(form_name=key, klass=klass)
                ) for key, klass in form_classes.items()
            ]
        )

    def get_initial(self, form_name):
        initial_method = 'get_{}_initial'.format(form_name)
        if hasattr(self, initial_method):
            return getattr(self, initial_method)()
        else:
            return self.initial.copy()

    def get_prefix(self, form_name):
        return self.prefixes.get(form_name, self.prefix)

    def post(self, request, *args, **kwargs):
        if all([form.is_valid() for form in self.forms.values()]):
            return self.forms_valid(forms=self.forms)
        else:
            return self.forms_invalid(forms=self.forms)


class AddRemoveView(
    ExternalObjectViewMixin, ExtraContextViewMixin, ViewPermissionCheckViewMixin,
    RestrictedQuerysetViewMixin, MultiFormView
):
    form_classes = {'form_available': ChoiceForm, 'form_added': ChoiceForm}
    list_added_help_text = _(
        'Select entries to be removed. Hold Control to select multiple '
        'entries. Once the selection is complete, click the button below '
        'or double click the list to activate the action.'
    )
    list_available_help_text = _(
        'Select entries to be added. Hold Control to select multiple '
        'entries. Once the selection is complete, click the button below '
        'or double click the list to activate the action.'
    )

    # Form titles
    list_added_title = None
    list_available_title = None

    # Attributes to filter the object to which selections will be added or
    # remove
    main_object_model = None
    main_object_permission = None
    main_object_pk_url_kwarg = None
    main_object_pk_url_kwargs = None
    main_object_source_queryset = None

    # Attributes to filter the queryset of the selection
    secondary_object_model = None
    secondary_object_permission = None
    secondary_object_source_queryset = None

    # Main object methods to use to add and remove selections
    main_object_method_add_name = None
    main_object_method_remove_name = None

    # If a method is not specified, use this related field to add and remove
    # selections
    related_field = None

    prefixes = {'form_available': 'available', 'form_added': 'added'}

    def __init__(self, *args, **kwargs):
        self.external_object_class = self.main_object_model
        self.external_object_permission = self.main_object_permission
        self.external_object_pk_url_kwarg = self.main_object_pk_url_kwarg
        self.external_object_pk_url_kwargs = self.main_object_pk_url_kwargs
        self.external_object_queryset = self.main_object_source_queryset

        super().__init__(*args, **kwargs)

    def _action_add(self, queryset):
        kwargs = {'queryset': queryset}
        kwargs.update(self.get_action_add_extra_kwargs())
        kwargs.update(self.get_actions_extra_kwargs())

        if hasattr(self, 'action_add'):
            self.action_add(**kwargs)
        elif self.main_object_method_add_name:
            getattr(self.main_object, self.main_object_method_add_name)(**kwargs)
        elif self.related_field:
            getattr(self.main_object, self.related_field).add(*queryset)
        else:
            raise ImproperlyConfigured(
                'View {} must be called with a main_object_method_add_name, a '
                'related_field, or an action_add '
                'method.'.format(self.__class__.__name__)
            )

    def _action_remove(self, queryset):
        kwargs = {'queryset': queryset}
        kwargs.update(self.get_action_remove_extra_kwargs())
        kwargs.update(self.get_actions_extra_kwargs())

        if hasattr(self, 'action_remove'):
            self.action_remove(**kwargs)
        elif self.main_object_method_remove_name:
            getattr(self.main_object, self.main_object_method_remove_name)(**kwargs)
        elif self.related_field:
            getattr(self.main_object, self.related_field).remove(*queryset)
        else:
            raise ImproperlyConfigured(
                'View {} must be called with a main_object_method_remove_name, a '
                'related_field, or an action_remove '
                'method.'.format(self.__class__.__name__)
            )

    def dispatch(self, request, *args, **kwargs):
        self.main_object = self.get_external_object()
        result = super().dispatch(request=request, *args, **kwargs)
        return result

    def forms_valid(self, forms):
        selection_add = None
        selection_remove = None

        if 'available-add_all' in self.request.POST:
            selection_add = self.get_list_available_queryset()
        else:
            if 'available-selection' in self.request.POST:
                selection_add = self.get_list_available_queryset().filter(
                    pk__in=forms['form_available'].cleaned_data['selection']
                )

        if selection_add:
            self._action_add(queryset=selection_add)

        if 'added-remove_all' in self.request.POST:
            selection_remove = self.get_list_added_queryset()
        else:
            if 'added-selection' in self.request.POST:
                selection_remove = self.get_list_added_queryset().filter(
                    pk__in=forms['form_added'].cleaned_data['selection']
                )

        if selection_remove:
            self._action_remove(queryset=selection_remove)

        return super().forms_valid(forms=forms)

    def generate_choices(self, queryset):
        for obj in queryset:
            yield (obj.pk, force_text(s=obj))

    def get_action_add_extra_kwargs(self):
        # Keyword arguments to apply to the add method
        return {}

    def get_action_remove_extra_kwargs(self):
        # Keyword arguments to apply to the remove method
        return {}

    def get_actions_extra_kwargs(self):
        # Keyword arguments to apply to both the add and remove methods
        return {}

    def get_context_data(self, **kwargs):
        # Use get_context_data to leave the get_extra_context for subclasses
        context = super().get_context_data(**kwargs)
        context.update(
            {
                'subtemplates_list': [
                    {
                        'name': 'appearance/generic_form_subtemplate.html',
                        'column_class': 'col-xs-12 col-sm-6 col-md-6 col-lg-6',
                        'context': {
                            'extra_buttons': [
                                {
                                    'label': _('Add all'),
                                    'icon': icon_add_all,
                                    'name': 'add_all',
                                }
                            ],
                            'form': self.forms['form_available'],
                            'form_css_classes': 'form-hotkey-double-click',
                            'hide_labels': True,
                            'submit_icon': icon_assign_remove_add,
                            'submit_label': _('Add'),
                            'title': self.list_available_title or ' ',
                        }
                    },
                    {
                        'name': 'appearance/generic_form_subtemplate.html',
                        'column_class': 'col-xs-12 col-sm-6 col-md-6 col-lg-6',
                        'context': {
                            'extra_buttons': [
                                {
                                    'label': _('Remove all'),
                                    'icon': icon_remove_all,
                                    'name': 'remove_all',
                                }
                            ],
                            'form': self.forms['form_added'],
                            'form_css_classes': 'form-hotkey-double-click',
                            'hide_labels': True,
                            'submit_icon': icon_assign_remove_remove,
                            'submit_label': _('Remove'),
                            'title': self.list_added_title or ' ',
                        }
                    }
                ]
            }
        )

        return context

    def get_disabled_choices(self):
        return ()

    def get_form_extra_kwargs(self, form_name):
        if form_name == 'form_available':
            return {
                'choices': self.generate_choices(
                    queryset=self.get_list_available_queryset()
                ),
                'help_text': self.get_list_available_help_text()
            }
        else:
            return {
                'choices': self.generate_choices(
                    queryset=self.get_list_added_queryset()
                ),
                'disabled_choices': self.get_disabled_choices(),
                'help_text': self.get_list_added_help_text()
            }

    def get_list_added_help_text(self):
        return self.list_added_help_text

    def get_list_added_queryset(self):
        if not self.related_field:
            raise ImproperlyConfigured(
                'View {} must be called with either a related_field or '
                'override .get_list_added_queryset().'.format(
                    self.__class__.__name__
                )
            )

        return self.get_secondary_object_list().filter(
            pk__in=getattr(self.main_object, self.related_field).values('pk')
        )

    def get_list_available_help_text(self):
        return self.list_available_help_text

    def get_list_available_queryset(self):
        return self.get_secondary_object_list().exclude(
            pk__in=self.get_list_added_queryset().values('pk')
        )

    def get_secondary_object_list(self):
        queryset = self.get_secondary_object_source_queryset()

        if queryset is None:
            queryset = self.secondary_object_model._meta.default_manager.all()

        if self.secondary_object_permission:
            return AccessControlList.objects.restrict_queryset(
                permission=self.secondary_object_permission, queryset=queryset,
                user=self.request.user
            )
        else:
            return queryset

    def get_secondary_object_source_queryset(self):
        return self.secondary_object_source_queryset

    def get_success_url(self):
        # Redirect to the same view
        return reverse(
            viewname=self.request.resolver_match.view_name,
            kwargs=self.request.resolver_match.kwargs
        )


class ConfirmView(
    RestrictedQuerysetViewMixin, ViewPermissionCheckViewMixin,
    ExtraContextViewMixin, RedirectionViewMixin, TemplateView
):
    """
    View that will execute an view action upon user Yes/No confirmation.
    """
    template_name = 'appearance/generic_confirm.html'

    def get_context_data(self, **kwargs):
        context = {
            'submit_icon': icon_confirm_form_submit,
            'cancel_icon': icon_confirm_form_cancel
        }
        context.update(super().get_context_data(**kwargs))
        return context

    def post(self, request, *args, **kwargs):
        self.view_action()
        return HttpResponseRedirect(redirect_to=self.get_success_url())


class FormView(
    ViewPermissionCheckViewMixin, ExtraContextViewMixin, RedirectionViewMixin,
    FormExtraKwargsViewMixin, DjangoFormView
):
    """
    Basic form view that will check for view level permission, allow
    providing extra context, extra keyword arguments for the forms, and
    customizable redirection.
    """
    template_name = 'appearance/generic_form.html'


class DynamicFormView(DynamicFormViewMixin, FormView):
    """Form view that uses a single dynamic form."""


class MultipleObjectFormActionView(
    ExtraContextViewMixin, ObjectActionViewMixin, ViewPermissionCheckViewMixin,
    RestrictedQuerysetViewMixin, MultipleObjectViewMixin, FormExtraKwargsViewMixin,
    RedirectionViewMixin, DjangoFormView
):
    """
    This view will present a form and upon receiving a POST request will
    perform an action on an object or queryset.
    """
    template_name = 'appearance/generic_form.html'

    def __init__(self, *args, **kwargs):
        result = super().__init__(*args, **kwargs)

        if self.__class__.mro()[0].get_queryset != MultipleObjectFormActionView.get_queryset:
            raise ImproperlyConfigured(
                '%(cls)s is overloading the get_queryset method. Subclasses '
                'should implement the get_source_queryset method instead. ' % {
                    'cls': self.__class__.__name__
                }
            )

        return result

    def form_invalid(self, form):
        if settings.DEBUG or settings.TESTING:
            logger.debug('Form invalid; %s', form.errors)

        return super().form_invalid(form=form)

    def form_valid(self, form):
        self.view_action(form=form)
        return super().form_valid(form=form)

    def get_queryset(self):
        try:
            return super().get_queryset()
        except ImproperlyConfigured:
            self.queryset = self.get_source_queryset()
            return super().get_queryset()


class MultipleObjectConfirmActionView(
    ExtraContextViewMixin, ObjectActionViewMixin,
    ViewPermissionCheckViewMixin, RestrictedQuerysetViewMixin,
    MultipleObjectViewMixin, RedirectionViewMixin, TemplateView
):
    """
    Form that will execute an action to a queryset upon user Yes/No
    confirmation.
    """
    template_name = 'appearance/generic_confirm.html'

    def __init__(self, *args, **kwargs):
        result = super().__init__(*args, **kwargs)

        if self.__class__.mro()[0].get_queryset != MultipleObjectConfirmActionView.get_queryset:
            raise ImproperlyConfigured(
                '%(cls)s is overloading the get_queryset method. Subclasses '
                'should implement the get_source_queryset method instead. ' % {
                    'cls': self.__class__.__name__
                }
            )

        return result

    def get_queryset(self):
        try:
            return super().get_queryset()
        except ImproperlyConfigured:
            self.queryset = self.get_source_queryset()
            return super().get_queryset()

    def post(self, request, *args, **kwargs):
        self.view_action()
        return HttpResponseRedirect(redirect_to=self.get_success_url())


class RelationshipView(FormView):
    form_class = None
    model = None
    model_permission = None
    pk_url_kwarg = None
    relationship_related_field = None
    relationship_related_query_field = None
    sub_model = None
    sub_model_permission = None

    def form_valid(self, forms):
        super_result = super().form_valid(form=forms)

        try:
            for form in forms:
                form.save()
        except Exception as exception:
            messages.error(
                message=_(
                    'Error updating relationship; %s'
                ) % exception, request=self.request
            )
            if settings.DEBUG or settings.TESTING:
                raise
        else:
            messages.success(
                message=_('Relationships updated successfully'),
                request=self.request
            )

        return super_result

    def get_object(self):
        queryset = AccessControlList.objects.restrict_queryset(
            permission=self.model_permission,
            queryset=self.model._meta.default_manager.all(),
            user=self.request.user
        )

        return get_object_or_404(
            klass=queryset, pk=self.kwargs[self.pk_url_kwarg]
        )

    def get_sub_model_queryset(self):
        queryset = self.sub_model.objects.all()
        return AccessControlList.objects.restrict_queryset(
            permission=self.sub_model_permission,
            user=self.request.user, queryset=queryset
        )


class SimpleView(ViewPermissionCheckViewMixin, ExtraContextViewMixin, TemplateView):
    """
    Basic template view class with permission check and extra context.
    """


class SingleObjectCreateView(
    ObjectNameViewMixin, ViewPermissionCheckViewMixin, ExtraContextViewMixin,
    RedirectionViewMixin, FormExtraKwargsViewMixin, CreateView
):
    error_message_duplicate = None
    template_name = 'appearance/generic_form.html'

    def form_valid(self, form):
        # This overrides the original Django form_valid method

        self.object = form.save(commit=False)

        if hasattr(self, 'get_instance_extra_data'):
            for key, value in self.get_instance_extra_data().items():
                setattr(self.object, key, value)

        if hasattr(self, 'get_save_extra_data'):
            save_extra_data = self.get_save_extra_data()
        else:
            save_extra_data = {}

        # Validate duplicates first
        try:
            self.object.validate_unique()
        except ValidationError as exception:
            context = self.get_context_data()

            error_message = self.get_error_message_duplicate() or _(
                'Duplicate data error: %(error)s'
            ) % {
                'error': '\n'.join(exception.messages)
            }

            messages.error(
                message=error_message, request=self.request
            )
            return super().form_invalid(form=form)

        try:
            self.object.save(**save_extra_data)
        except Exception as exception:
            if settings.DEBUG or settings.TESTING:
                raise
            else:
                context = self.get_context_data()

                messages.error(
                    message=_('%(object)s not created, error: %(error)s') % {
                        'object': self.get_object_name(context=context),
                        'error': exception
                    }, request=self.request
                )
                return super().form_invalid(form=form)
        else:
            context = self.get_context_data()

            messages.success(
                message=_(
                    '%(object)s created successfully.'
                ) % {'object': self.get_object_name(context=context)},
                request=self.request
            )

        return HttpResponseRedirect(redirect_to=self.get_success_url())

    def get_error_message_duplicate(self):
        return self.error_message_duplicate


class SingleObjectDeleteView(
    ObjectNameViewMixin, ExtraDataDeleteViewMixin, ViewPermissionCheckViewMixin,
    RestrictedQuerysetViewMixin, ExtraContextViewMixin, RedirectionViewMixin, DeleteView
):
    template_name = 'appearance/generic_confirm.html'

    def __init__(self, *args, **kwargs):
        result = super().__init__(*args, **kwargs)

        if self.__class__.mro()[0].get_queryset != SingleObjectDeleteView.get_queryset:
            raise ImproperlyConfigured(
                '%(cls)s is overloading the get_queryset method. Subclasses '
                'should implement the get_source_queryset method instead. ' % {
                    'cls': self.__class__.__name__
                }
            )

        return result

    def delete(self, request, *args, **kwargs):
        self.object = self.get_object()
        context = self.get_context_data()
        object_name = self.get_object_name(context=context)

        try:
            result = super().delete(request, *args, **kwargs)
        except Exception as exception:
            messages.error(
                message=_('%(object)s not deleted, error: %(error)s.') % {
                    'object': object_name,
                    'error': exception
                }, request=self.request
            )
            raise
        else:
            messages.success(
                message=_(
                    '%(object)s deleted successfully.'
                ) % {'object': object_name},
                request=self.request
            )

            return result

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({'delete_view': True})
        return context

    def get_queryset(self):
        try:
            return super().get_queryset()
        except ImproperlyConfigured:
            self.queryset = self.get_source_queryset()
            return super().get_queryset()


class SingleObjectDetailView(
    ObjectNameViewMixin, ViewPermissionCheckViewMixin,
    RestrictedQuerysetViewMixin, FormExtraKwargsViewMixin,
    ExtraContextViewMixin, ModelFormMixin, DetailView
):
    template_name = 'appearance/generic_form.html'

    def __init__(self, *args, **kwargs):
        result = super().__init__(*args, **kwargs)

        if self.__class__.mro()[0].get_queryset != SingleObjectDetailView.get_queryset:
            raise ImproperlyConfigured(
                '%(cls)s is overloading the get_queryset method. Subclasses '
                'should implement the get_source_queryset method instead. ' % {
                    'cls': self.__class__.__name__
                }
            )

        return result

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({'read_only': True, 'form': self.get_form()})
        return context

    def get_form(self):
        try:
            return super().get_form()
        except Exception as exception:
            if settings.DEBUG or settings.TESTING:
                raise
            else:
                messages.error(
                    message=_(
                        'Error retrieving %(object)s; %(error)s'
                    ) % {
                        'object': self.get_object_name(
                            context={'object': self.object}
                        ),
                        'error': exception
                    }, request=self.request
                )

    def get_queryset(self):
        try:
            return super().get_queryset()
        except ImproperlyConfigured:
            self.queryset = self.get_source_queryset()
            return super().get_queryset()


class BaseDownloadView(DownloadViewMixin, ViewPermissionCheckViewMixin, View):
    def get(self, request, *args, **kwargs):
        return self.render_to_response()


class SingleObjectDownloadView(
    RestrictedQuerysetViewMixin, SingleObjectMixin, BaseDownloadView
):
    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        return super().get(
            request, *args, **kwargs
        )

    def get_download_file_object(self):
        return self.object.open()

    def get_download_label(self):
        return force_text(s=self.object)


class MultipleObjectDownloadView(
    RestrictedQuerysetViewMixin, MultipleObjectViewMixin, BaseDownloadView
):
    """
    View that support receiving multiple objects via a pk_list query.
    """
    def __init__(self, *args, **kwargs):
        result = super().__init__(*args, **kwargs)

        if self.__class__.mro()[0].get_queryset != MultipleObjectDownloadView.get_queryset:
            raise ImproperlyConfigured(
                '%(cls)s is overloading the get_queryset method. Subclasses '
                'should implement the get_source_queryset method instead. ' % {
                    'cls': self.__class__.__name__
                }
            )

        return result

    def get_queryset(self):
        try:
            return super().get_queryset()
        except ImproperlyConfigured:
            self.queryset = self.get_source_queryset()
            return super().get_queryset()


class SingleObjectDynamicFormCreateView(
    DynamicFormViewMixin, SingleObjectCreateView
):
    """
    A form that will allow creation of a single instance from the values
    of a dynamic field form.
    """


class SingleObjectEditView(
    ObjectNameViewMixin, ViewPermissionCheckViewMixin, RestrictedQuerysetViewMixin,
    ExtraContextViewMixin, FormExtraKwargsViewMixin, RedirectionViewMixin, UpdateView
):
    template_name = 'appearance/generic_form.html'

    def form_valid(self, form):
        # This overrides the original Django form_valid method

        self.object = form.save(commit=False)

        if hasattr(self, 'get_instance_extra_data'):
            for key, value in self.get_instance_extra_data().items():
                setattr(self.object, key, value)

        if hasattr(self, 'get_save_extra_data'):
            save_extra_data = self.get_save_extra_data()
        else:
            save_extra_data = {}

        context = self.get_context_data()
        object_name = self.get_object_name(context=context)

        try:
            self.object.save(**save_extra_data)
        except Exception as exception:
            if settings.DEBUG or settings.TESTING:
                raise
            else:
                messages.error(
                    message=_('%(object)s not updated, error: %(error)s.') % {
                        'object': object_name,
                        'error': exception
                    }, request=self.request
                )
                return super().form_invalid(form=form)
        else:
            messages.success(
                message=_(
                    '%(object)s updated successfully.'
                ) % {'object': object_name}, request=self.request
            )

        return HttpResponseRedirect(redirect_to=self.get_success_url())

    def get_object(self, queryset=None):
        obj = super().get_object(queryset=queryset)

        if hasattr(self, 'get_instance_extra_data'):
            for key, value in self.get_instance_extra_data().items():
                setattr(obj, key, value)

        return obj


class SingleObjectDynamicFormEditView(
    DynamicFormViewMixin, SingleObjectEditView
):
    """
    A form that will allow editing a single instance from the values
    of a dynamic field form.
    """


class SingleObjectListView(
    SortingViewMixin, ListModeViewMixin, PaginationMixin,
    ViewPermissionCheckViewMixin, RestrictedQuerysetViewMixin,
    ExtraContextViewMixin, RedirectionViewMixin, ListView
):
    """
    A view that will generate a list of instances from a queryset.
    """
    template_name = 'appearance/generic_list.html'

    def __init__(self, *args, **kwargs):
        result = super().__init__(*args, **kwargs)

        if self.__class__.mro()[0].get_queryset != SingleObjectListView.get_queryset:
            raise ImproperlyConfigured(
                '%(cls)s is overloading the get_queryset method. Subclasses '
                'should implement the get_source_queryset method instead. ' % {
                    'cls': self.__class__.__name__
                }
            )

        return result

    def get_paginate_by(self, queryset):
        return setting_paginate_by.value

    def get_queryset(self):
        try:
            queryset = super().get_queryset()
        except ImproperlyConfigured:
            self.queryset = self.get_source_queryset()
            queryset = super().get_queryset()

        return queryset


class MultipleObjectDeleteView(MultipleObjectConfirmActionView):
    success_message_single = _('"%(object)s" deleted successfully.')
    success_message_singular = _('%(count)d object deleted successfully.')
    success_message_plural = _('%(count)d objects deleted successfully.')
    title_single = _('Delete "%(object)s".')
    title_singular = _('Delete %(count)d object.')
    title_plural = _('Delete %(count)d objects.')

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            {
                'delete_view': True,
            }
        )

        if self.object_list.count() == 1:
            context.update(
                {
                    'object': self.object_list.first(),
                }
            )

        return context

    def object_action(self, instance, form=None):
        if hasattr(self, 'get_instance_extra_data'):
            for key, value in self.get_instance_extra_data().items():
                setattr(instance, key, value)

        if hasattr(self, 'get_delete_extra_data'):
            instance.delete(**self.get_delete_extra_data())
        else:
            instance.delete()



import logging

from django.utils.module_loading import import_string

from ..settings import setting_backend, setting_default_lock_timeout

logger = logging.getLogger(name=__name__)


class LockingBackend:
    """
    Base class for the lock backends. Defines the base methods that each
    subclass must define.
    """
    _is_initialized = False

    @classmethod
    def _initialize(cls):
        """
        Optional class method for subclasses to overload.
        """
        return

    @staticmethod
    def get_backend():
        return import_string(dotted_path=setting_backend.value)

    @classmethod
    def acquire_lock(cls, name, timeout=None):
        timeout = timeout or setting_default_lock_timeout.value
        logger.debug('acquiring lock: %s, timeout: %s', name, timeout)
        return cls._acquire_lock(name=name, timeout=timeout)

    @classmethod
    def purge_locks(cls):
        if not cls._is_initialized:
            cls._initialize()
            cls._is_initialized = True

        logger.debug('purging locks')
        return cls._purge_locks()

    def __init__(self, *args, **kwargs):
        if not self.__class__._is_initialized:
            self.__class__._initialize()
            self.__class__._is_initialized = True

        return self._init(*args, **kwargs)

    def release(self):
        logger.debug('releasing lock: %s', self.name)
        return self._release()



from django.db.models import Q

from ...models import (
    Workflow, WorkflowRuntimeProxy, WorkflowStateRuntimeProxy
)
from ...tasks import (
    task_launch_all_workflows, task_launch_all_workflow_for,
    task_launch_workflow, task_launch_workflow_for
)

from ..literals import (
    TEST_WORKFLOW_INSTANCE_LOG_ENTRY_COMMENT,
    TEST_WORKFLOW_TEMPLATE_INTERNAL_NAME, TEST_WORKFLOW_TEMPLATE_LABEL,
    TEST_WORKFLOW_TEMPLATE_LABEL_EDITED,
    TEST_WORKFLOW_TEMPLATE_STATE_COMPLETION,
    TEST_WORKFLOW_TEMPLATE_STATE_LABEL,
    TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL
)


class DocumentTypeAddRemoveWorkflowTemplateViewTestMixin:
    def _request_test_document_type_workflow_template_add_remove_get_view(self):
        return self.get(
            viewname='document_states:document_type_workflow_templates',
            kwargs={
                'document_type_id': self.test_document_type.pk,
            }
        )

    def _request_test_document_type_workflow_template_add_view(self):
        return self.post(
            viewname='document_states:document_type_workflow_templates',
            kwargs={
                'document_type_id': self.test_document_type.pk,
            }, data={
                'available-submit': 'true',
                'available-selection': self.test_workflow_template.pk
            }
        )

    def _request_test_document_type_workflow_template_remove_view(self):
        return self.post(
            viewname='document_states:document_type_workflow_templates',
            kwargs={
                'document_type_id': self.test_document_type.pk,
            }, data={
                'added-submit': 'true',
                'added-selection': self.test_workflow_template.pk
            }
        )


class WorkflowTaskTestCaseMixin:
    def _execute_task_launch_all_workflows(self):
        task_launch_all_workflows.apply_async().get()

    def _execute_task_launch_all_workflow_for(self):
        task_launch_all_workflow_for.apply_async(
            kwargs={
                'document_id': self.test_document.pk,
            }
        ).get()

    def _execute_task_launch_workflow(self):
        task_launch_workflow.apply_async(
            kwargs={
                'workflow_id': self.test_workflow_template.pk
            }
        ).get()

    def _execute_task_launch_workflow_for(self):
        task_launch_workflow_for.apply_async(
            kwargs={
                'document_id': self.test_document.pk,
                'workflow_id': self.test_workflow_template.pk
            }
        ).get()


class WorkflowTemplateDocumentTypeViewTestMixin:
    def _request_test_workflow_template_document_type_add_remove_get_view(self):
        return self.get(
            viewname='document_states:workflow_template_document_types',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }
        )

    def _request_test_workflow_template_document_type_add_view(self):
        return self.post(
            viewname='document_states:workflow_template_document_types',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
            }, data={
                'available-submit': 'true',
                'available-selection': self.test_document_type.pk
            }
        )

    def _request_test_workflow_template_document_type_remove_view(self):
        return self.post(
            viewname='document_states:workflow_template_document_types',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
            }, data={
                'added-submit': 'true',
                'added-selection': self.test_document_type.pk
            }
        )


class WorkflowTemplateAPIViewTestMixin:
    def _request_test_workflow_template_create_api_view(self, extra_data=None):
        data = {
            'internal_name': TEST_WORKFLOW_TEMPLATE_INTERNAL_NAME,
            'label': TEST_WORKFLOW_TEMPLATE_LABEL,
        }

        if extra_data:
            data.update(extra_data)

        pk_list = list(Workflow.objects.values('pk'))

        response = self.post(
            viewname='rest_api:workflow-template-list', data=data
        )

        try:
            self.test_workflow_template = Workflow.objects.get(
                ~Q(pk__in=pk_list)
            )
        except Workflow.DoesNotExist:
            self.test_workflow_template = None

        return response

    def _request_test_workflow_template_delete_api_view(self):
        return self.delete(
            viewname='rest_api:workflow-template-detail', kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }
        )

    def _request_test_workflow_template_detail_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-detail', kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }
        )

    def _request_test_workflow_template_edit_via_patch_view(self):
        return self.patch(
            viewname='rest_api:workflow-template-detail', kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }, data={'label': TEST_WORKFLOW_TEMPLATE_LABEL_EDITED}
        )

    def _request_test_workflow_template_edit_via_put_view(self):
        return self.put(
            viewname='rest_api:workflow-template-detail', kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }, data={
                'internal_name': TEST_WORKFLOW_TEMPLATE_INTERNAL_NAME,
                'label': TEST_WORKFLOW_TEMPLATE_LABEL_EDITED
            }
        )

    def _request_test_workflow_template_image_view_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-image', kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }
        )

    def _request_test_workflow_template_list_api_view(self):
        return self.get(viewname='rest_api:workflow-template-list')


class WorkflowTemplateDocumentTypeAPIViewMixin:
    def _request_test_workflow_template_document_type_add_api_view(self):
        return self.post(
            viewname='rest_api:workflow-template-document-type-add',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
            }, data={
                'document_type_id': self.test_document_type.pk
            }
        )

    def _request_test_workflow_template_document_type_list_api_view(self):
        return self.get(
            viewname='rest_api:workflow-template-document-type-list', kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }
        )

    def _request_test_workflow_template_document_type_remove_api_view(self):
        return self.post(
            viewname='rest_api:workflow-template-document-type-remove',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
            }, data={
                'document_type_id': self.test_document_type.pk
            }
        )


class WorkflowTemplateTestMixin:
    def setUp(self):
        super().setUp()
        self.test_workflow_runtime_proxies = []
        self.test_workflow_template_state_runtime_proxies = []
        self.test_workflow_template_states = []
        self.test_workflow_template_transitions = []
        self.test_workflow_templates = []

    def _create_test_workflow_template(
        self, add_test_document_type=False, auto_launch=True
    ):
        total_test_workflow_templates = len(self.test_workflow_templates)
        label = '{}_{}'.format(
            TEST_WORKFLOW_TEMPLATE_LABEL, total_test_workflow_templates
        )
        internal_name = '{}_{}'.format(
            TEST_WORKFLOW_TEMPLATE_INTERNAL_NAME,
            total_test_workflow_templates
        )

        self.test_workflow_template = Workflow.objects.create(
            auto_launch=auto_launch, label=label,
            internal_name=internal_name
        )
        self.test_workflow_templates.append(
            self.test_workflow_template
        )
        self.test_workflow_runtime_proxy = WorkflowRuntimeProxy.objects.get(
            pk=self.test_workflow_template.pk
        )
        self.test_workflow_runtime_proxies.append(
            self.test_workflow_runtime_proxy
        )

        if add_test_document_type:
            self.test_workflow_template.document_types.add(
                self.test_document_type
            )

    def _create_test_workflow_template_state(self):
        total_test_workflow_template_states = len(
            self.test_workflow_template_states
        )
        label = '{}_{}'.format(
            TEST_WORKFLOW_TEMPLATE_STATE_LABEL,
            total_test_workflow_template_states
        )
        initial = self.test_workflow_template.states.count() == 0

        self.test_workflow_template_state = self.test_workflow_template.states.create(
            completion=TEST_WORKFLOW_TEMPLATE_STATE_COMPLETION, initial=initial,
            label=label
        )
        self.test_workflow_template_states.append(
            self.test_workflow_template_state
        )
        self.test_workflow_template_state_runtime_proxy = WorkflowStateRuntimeProxy.objects.get(
            pk=self.test_workflow_template_state.pk
        )
        self.test_workflow_template_state_runtime_proxies.append(
            self.test_workflow_template_state_runtime_proxy
        )

    def _create_test_workflow_template_transition(self):
        total_test_workflow_template_transitions = len(
            self.test_workflow_template_transitions
        )
        label = '{}_{}'.format(
            TEST_WORKFLOW_TEMPLATE_TRANSITION_LABEL,
            total_test_workflow_template_transitions
        )

        self.test_workflow_template_transition = self.test_workflow_template.transitions.create(
            label=label,
            origin_state=self.test_workflow_template_states[0],
            destination_state=self.test_workflow_template_states[1],
        )

        self.test_workflow_template_transitions.append(
            self.test_workflow_template_transition
        )

    def _create_test_workflow_template_instance_log_entry(self):
        self.test_document.workflows.first().log_entries.create(
            comment=TEST_WORKFLOW_INSTANCE_LOG_ENTRY_COMMENT,
            transition=self.test_workflow_template_transition,
            user=self._test_case_user
        )

    def _transition_test_workflow_instance(self, extra_data=None):
        self.test_document.workflows.first().do_transition(
            comment=TEST_WORKFLOW_INSTANCE_LOG_ENTRY_COMMENT,
            extra_data=extra_data,
            transition=self.test_workflow_template_transition
        )


class WorkflowToolViewTestMixin:
    def _request_workflow_launch_view(self):
        return self.post(
            viewname='document_states:tool_launch_workflows',
        )


class WorkflowTemplateViewTestMixin:
    def _request_test_workflow_template_create_view(self):
        data = {
            'internal_name': TEST_WORKFLOW_TEMPLATE_INTERNAL_NAME,
            'label': TEST_WORKFLOW_TEMPLATE_LABEL,
        }

        pk_list = list(Workflow.objects.values('pk'))

        response = self.post(
            viewname='document_states:workflow_template_create', data=data
        )

        try:
            self.test_workflow_template = Workflow.objects.get(
                ~Q(pk__in=pk_list)
            )
        except Workflow.DoesNotExist:
            self.test_workflow_template = None

        return response

    def _request_test_workflow_template_delete_view(self):
        return self.post(
            viewname='document_states:workflow_template_single_delete',
            kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }
        )

    def _request_test_workflow_template_edit_view(self):
        return self.post(
            viewname='document_states:workflow_template_edit', kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
            }, data={
                'auto_launch': True,
                'label': TEST_WORKFLOW_TEMPLATE_LABEL_EDITED,
                'internal_name': self.test_workflow_template.internal_name
            }
        )

    def _request_test_workflow_template_launch_view(self):
        return self.post(
            viewname='document_states:workflow_template_launch', kwargs={
                'workflow_template_id': self.test_workflow_template.pk
            }
        )

    def _request_test_workflow_template_list_view(self):
        return self.get(
            viewname='document_states:workflow_template_list',
        )

    def _request_test_workflow_template_preview_view(self):
        return self.get(
            viewname='document_states:workflow_template_preview', kwargs={
                'workflow_template_id': self.test_workflow_template.pk,
            }
        )



from ..classes import MailerBackend


class TestBackend(MailerBackend):
    """
    User mailer backend to use with tests
    """
    class_path = 'django.core.mail.backends.locmem.EmailBackend'
    label = 'Django local memory backend'



from django import forms

from ..fields import DocumentFilePageField

__all__ = ('DocumentFilePageForm',)


class DocumentFilePageForm(forms.Form):
    document_file_page = DocumentFilePageField()

    def __init__(self, *args, **kwargs):
        instance = kwargs.pop('instance', None)
        rotation = kwargs.pop('rotation', None)
        zoom = kwargs.pop('zoom', None)
        super().__init__(*args, **kwargs)
        self.fields['document_file_page'].initial = instance
        self.fields['document_file_page'].widget.attrs.update({
            'zoom': zoom,
            'rotation': rotation,
        })



from django.conf.urls import url

from .api_views import APITemplateDetailView, APITemplateListView

from .views import DocumentTemplateSandboxView

urlpatterns = [
    url(
        regex=r'^documents/(?P<document_id>\d+)/sandbox/$',
        name='document_template_sandbox',
        view=DocumentTemplateSandboxView.as_view()
    ),
]

api_urls = [
    url(
        regex=r'^templates/$', view=APITemplateListView.as_view(),
        name='template-list'
    ),
    url(
        regex=r'^templates/(?P<name>[-\w]+)/$',
        view=APITemplateDetailView.as_view(), name='template-detail'
    )
]



import logging

from django.conf import settings
from django.http import HttpResponse
from django.views.decorators.cache import cache_control, patch_cache_control

from rest_framework import status

from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ActionAPIViewMixin

from ..literals import DOCUMENT_IMAGE_TASK_TIMEOUT
from ..permissions import (
    permission_document_version_create, permission_document_version_delete,
    permission_document_version_edit, permission_document_version_export,
    permission_document_version_view
)
from ..serializers.document_version_serializers import (
    DocumentVersionSerializer, DocumentVersionPageSerializer
)
from ..settings import setting_document_version_page_image_cache_time
from ..tasks import (
    task_document_version_export, task_document_version_page_image_generate
)

from .mixins import (
    ParentObjectDocumentAPIViewMixin, ParentObjectDocumentVersionAPIViewMixin
)

logger = logging.getLogger(name=__name__)


class APIDocumentVersionDetailView(
    ParentObjectDocumentAPIViewMixin, generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected document version.
    get: Returns the selected document version details.
    patch: Edit the properties of the selected document version.
    put: Edit the properties of the selected document version.
    """
    lookup_url_kwarg = 'document_version_id'
    mayan_object_permissions = {
        'DELETE': (permission_document_version_delete,),
        'GET': (permission_document_version_view,),
        'PATCH': (permission_document_version_edit,),
        'PUT': (permission_document_version_edit,)
    }
    serializer_class = DocumentVersionSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
        }

    def get_queryset(self):
        return self.get_document().versions.all()


class APIDocumentVersionExportView(
    ActionAPIViewMixin, ParentObjectDocumentAPIViewMixin,
    generics.GenericAPIView
):
    """
    post: Exports the specified document version.
    """
    action_response_status = status.HTTP_202_ACCEPTED
    lookup_url_kwarg = 'document_version_id'
    mayan_object_permissions = {
        'POST': (permission_document_version_export,),
    }

    def get_queryset(self):
        return self.get_document().versions.all()

    def get_serializer(self):
        return None

    def perform_view_action(self):
        task_document_version_export.apply_async(
            kwargs={'document_version_id': self.get_object().pk}
        )


class APIDocumentVersionListView(
    ParentObjectDocumentAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Return a list of the selected document's versions.
    post: Create a new document version.
    """
    ordering_fields = ('active', 'comment')
    serializer_class = DocumentVersionSerializer

    def get_instance_extra_data(self):
        # This method is only called during POST, therefore filter only by
        # edit permission.
        return {
            '_event_actor': self.request.user,
            'document': self.get_document(
                permission=permission_document_version_create
            )
        }

    def get_queryset(self):
        # This method is only called during GET, therefore filter only by
        # the view permission.
        return self.get_document(
            permission=permission_document_version_view
        ).versions.all()


class APIDocumentVersionPageDetailView(
    ParentObjectDocumentVersionAPIViewMixin,
    generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected document version page.
    get: Returns the selected document version page details.
    patch: Edit the properties of the selected document version page.
    put: Edit the properties of the selected document version page.
    """
    lookup_url_kwarg = 'document_version_page_id'
    mayan_object_permissions = {
        'DELETE': (permission_document_version_edit,),
        'GET': (permission_document_version_view,),
        'PATCH': (permission_document_version_edit,),
        'PUT': (permission_document_version_edit,)
    }
    serializer_class = DocumentVersionPageSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_queryset(self):
        return self.get_document_version().pages.all()


class APIDocumentVersionPageImageView(
    ParentObjectDocumentVersionAPIViewMixin, generics.RetrieveAPIView
):
    """
    get: Returns an image representation of the selected document version page.
    """
    lookup_url_kwarg = 'document_version_page_id'
    mayan_object_permissions = {
        'GET': (permission_document_version_view,),
    }

    def get_queryset(self):
        return self.get_document_version().pages.all()

    def get_serializer(self, *args, **kwargs):
        return None

    def get_serializer_class(self):
        return None

    @cache_control(private=True)
    def retrieve(self, request, *args, **kwargs):
        width = request.GET.get('width')
        height = request.GET.get('height')
        zoom = request.GET.get('zoom')

        if zoom:
            zoom = int(zoom)

        rotation = request.GET.get('rotation')

        if rotation:
            rotation = int(rotation)

        maximum_layer_order = request.GET.get('maximum_layer_order')
        if maximum_layer_order:
            maximum_layer_order = int(maximum_layer_order)

        task = task_document_version_page_image_generate.apply_async(
            kwargs=dict(
                document_version_page_id=self.get_object().pk, width=width,
                height=height, zoom=zoom, rotation=rotation,
                maximum_layer_order=maximum_layer_order,
                user_id=request.user.pk
            )
        )

        kwargs = {'timeout': DOCUMENT_IMAGE_TASK_TIMEOUT}
        if settings.DEBUG:
            # In debug more, task are run synchronously, causing this method
            # to be called inside another task. Disable the check of nested
            # tasks when using debug mode.
            kwargs['disable_sync_subtasks'] = False

        cache_filename = task.get(**kwargs)
        cache_file = self.get_object().cache_partition.get_file(filename=cache_filename)
        with cache_file.open() as file_object:
            response = HttpResponse(file_object.read(), content_type='image')
            if '_hash' in request.GET:
                patch_cache_control(
                    response=response,
                    max_age=setting_document_version_page_image_cache_time.value
                )
            return response


class APIDocumentVersionPageListView(
    ParentObjectDocumentVersionAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Returns an list of the pages for the selected document version.
    post: Create a new document version page.
    """
    lookup_url_kwarg = 'document_version_page_id'
    serializer_class = DocumentVersionPageSerializer

    def get_instance_extra_data(self):
        # This method is only called during POST, therefore filter only by
        # edit permission.
        return {
            '_event_actor': self.request.user,
            'document_version': self.get_document_version(
                permission=permission_document_version_edit
            )
        }

    def get_queryset(self):
        # This method is only called during GET, therefore filter only by
        # the view permission.
        return self.get_document_version(
            permission=permission_document_version_view
        ).pages.all()



from datetime import timedelta

from ..models.document_models import Document
from ..settings import setting_stub_expiration_interval

from .base import GenericDocumentTestCase
from .literals import (
    TEST_SMALL_DOCUMENT_CHECKSUM, TEST_SMALL_DOCUMENT_FILENAME,
    TEST_SMALL_DOCUMENT_MIMETYPE, TEST_SMALL_DOCUMENT_SIZE
)


class DocumentTestCase(GenericDocumentTestCase):
    def test_document_creation(self):
        self.assertEqual(
            self.test_document.document_type.label,
            self.test_document_type.label
        )

        self.assertEqual(self.test_document.file_latest.exists(), True)
        self.assertEqual(
            self.test_document.file_latest.size,
            TEST_SMALL_DOCUMENT_SIZE
        )

        self.assertEqual(
            self.test_document.file_latest.mimetype,
            TEST_SMALL_DOCUMENT_MIMETYPE
        )
        self.assertEqual(
            self.test_document.file_latest.encoding, 'binary'
        )
        self.assertEqual(
            self.test_document.file_latest.checksum,
            TEST_SMALL_DOCUMENT_CHECKSUM
        )
        self.assertEqual(self.test_document.file_latest.pages.count(), 1)
        self.assertEqual(
            self.test_document.label, TEST_SMALL_DOCUMENT_FILENAME
        )

    def test_method_get_absolute_url(self):
        self._upload_test_document()

        self.assertTrue(self.test_document.get_absolute_url())


class DocumentManagerTestCase(GenericDocumentTestCase):
    auto_upload_test_document = False

    def test_document_stubs_deletion(self):
        document_stub = Document.objects.create(
            document_type=self.test_document_type
        )

        Document.objects.delete_stubs()

        self.assertEqual(Document.objects.count(), 1)

        document_stub.datetime_created = document_stub.datetime_created - timedelta(
            seconds=setting_stub_expiration_interval.value + 1
        )
        document_stub.save()

        Document.objects.delete_stubs()

        self.assertEqual(Document.objects.count(), 0)



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('events', '0001_initial'),
        ('document_states', '0004_workflow_internal_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='WorkflowTransitionTriggerEvent',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'stored_event_type', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='trigger_events', to='events.EventType',
                        verbose_name='Event type'
                    )
                ),
            ],
            options={
                'verbose_name': 'Workflow transition trigger event',
                'verbose_name_plural': 'Workflow transitions trigger events',
            },
        ),
        migrations.AddField(
            model_name='workflowtransition',
            name='trigger_time_period',
            field=models.PositiveIntegerField(
                blank=True, help_text='Amount of time after which this '
                'transition will trigger on its own.', null=True,
                verbose_name='Trigger time period'
            ),
        ),
        migrations.AddField(
            model_name='workflowtransition',
            name='trigger_time_unit',
            field=models.CharField(
                blank=True, choices=[
                    ('days', 'Days'), ('hours', 'Hours'),
                    ('minutes', 'Minutes')
                ], max_length=8, null=True, verbose_name='Trigger time unit'
            ),
        ),
        migrations.AddField(
            model_name='workflowtransitiontriggerevent',
            name='transition',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to='document_states.WorkflowTransition',
                verbose_name='Transition'
            ),
        ),
    ]



import logging

from django.utils.translation import ugettext_lazy as _

from mayan.apps.views.forms import FilteredSelectionForm

logger = logging.getLogger(name=__name__)


class CabinetListForm(FilteredSelectionForm):
    class Meta:
        allow_multiple = True
        field_name = 'cabinets'
        label = _('Cabinets')
        required = False
        widget_attributes = {'class': 'select2'}



TEST_EVENT_TYPE_NAMESPACE_LABEL = 'test event type namespace label'
TEST_EVENT_TYPE_NAMESPACE_NAME = 'test_event_type_namespace_name'
TEST_EVENT_TYPE_LABEL = 'test event type label'
TEST_EVENT_TYPE_NAME = 'test_event_type_name'



import os

from django.conf import settings
from django.utils.translation import ugettext_lazy as _

DEFAULT_GRAPHVIZ_DOT_PATH = '/usr/bin/dot'
DEFAULT_HTTP_ACTION_TIMEOUT = 4  # 4 seconds
DEFAULT_WORKFLOWS_IMAGE_CACHE_MAXIMUM_SIZE = 50 * 2 ** 20  # 50 Megabytes
DEFAULT_WORKFLOWS_IMAGE_CACHE_STORAGE_BACKEND = 'django.core.files.storage.FileSystemStorage'
DEFAULT_WORKFLOWS_IMAGE_CACHE_STORAGE_BACKEND_ARGUMENTS = {
    'location': os.path.join(settings.MEDIA_ROOT, 'workflows')
}
DEFAULT_WORKFLOWS_IMAGE_CACHE_TIME = '31556926'

FIELD_TYPE_CHOICE_CHAR = 1
FIELD_TYPE_CHOICE_INTEGER = 2
FIELD_TYPE_CHOICES = (
    (FIELD_TYPE_CHOICE_CHAR, _('Character')),
    (FIELD_TYPE_CHOICE_INTEGER, _('Number (Integer)')),
)
FIELD_TYPE_MAPPING = {
    FIELD_TYPE_CHOICE_CHAR: 'django.forms.CharField',
    FIELD_TYPE_CHOICE_INTEGER: 'django.forms.IntegerField',
}

STORAGE_NAME_WORKFLOW_CACHE = 'document_states__workflowimagecache'

SYMBOL_MATH_CONDITIONAL = '&rarr;'

TASK_GENERATE_WORKFLOW_IMAGE_RETRY_DELAY = 10

WIDGET_CLASS_TEXTAREA = 1
WIDGET_CLASS_CHOICES = (
    (WIDGET_CLASS_TEXTAREA, _('Text area')),
)
WIDGET_CLASS_MAPPING = {
    WIDGET_CLASS_TEXTAREA: 'django.forms.widgets.Textarea',
}

WORKFLOW_ACTION_ON_ENTRY = 1
WORKFLOW_ACTION_ON_EXIT = 2
WORKFLOW_ACTION_WHEN_CHOICES = (
    (WORKFLOW_ACTION_ON_ENTRY, _('On entry')),
    (WORKFLOW_ACTION_ON_EXIT, _('On exit')),
)
WORKFLOW_IMAGE_TASK_TIMEOUT = 60



import logging
from pathlib import Path

import whoosh
from whoosh import qparser
from whoosh.filedb.filestore import FileStorage
from whoosh.index import EmptyIndexError

from django.conf import settings

from mayan.apps.lock_manager.backends.base import LockingBackend
from mayan.apps.lock_manager.exceptions import LockError

from ..classes import SearchBackend, SearchField, SearchModel
from ..settings import setting_results_limit

from .literals import DJANGO_TO_WHOOSH_FIELD_MAP, WHOOSH_INDEX_DIRECTORY_NAME
logger = logging.getLogger(name=__name__)


class WhooshSearchBackend(SearchBackend):
    _resolved_field_maps = {}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.index_path = Path(
            self.kwargs.get(
                'index_path', Path(settings.MEDIA_ROOT, WHOOSH_INDEX_DIRECTORY_NAME)
            )
        )
        self.index_path.mkdir(exist_ok=True)

    def _search(self, query_string, search_model, user, global_and_search=False):
        index = self.get_index(search_model=search_model)

        id_list = []
        with index.searcher() as searcher:
            search_string = []

            if 'q' in query_string:
                # Emulate full field set search
                for search_field in self.get_search_model_fields(search_model=search_model):
                    search_string.append(
                        '{}:({})'.format(search_field.get_full_name(), query_string['q'])
                    )
            else:
                for key, value in query_string.items():
                    if value:
                        search_string.append(
                            '{}:({})'.format(key, value)
                        )

            global_logic_string = ' AND ' if global_and_search else ' OR '
            search_string = global_logic_string.join(search_string)

            logger.debug('search_string: %s', search_string)

            parser = qparser.QueryParser(
                fieldname='_', schema=index.schema
            )
            parser.remove_plugin_class(cls=qparser.WildcardPlugin)
            parser.add_plugin(pin=qparser.PrefixPlugin())
            query = parser.parse(text=search_string)
            results = searcher.search(
                q=query, limit=setting_results_limit.value
            )

            logger.debug('results: %s', results)

            for result in results:
                id_list.append(result['id'])

        queryset = search_model.get_queryset().filter(
            id__in=id_list
        ).distinct()

        return SearchBackend.limit_queryset(queryset=queryset)

    def clear_search_model_index(self, search_model):
        index = self.get_index(search_model=search_model)

        # Clear the model index
        self.get_storage().create_index(
            index.schema, indexname=search_model.get_full_name()
        )

    def deindex_instance(self, instance):
        try:
            lock = LockingBackend.get_backend().acquire_lock(
                name='dynamic_search_whoosh_deindex_instance'
            )
        except LockError:
            raise
        else:
            try:
                search_model = SearchModel.get_for_model(instance=instance)
                index = self.get_index(search_model=search_model)

                writer = index.writer()
                writer.delete_by_term('id', str(instance.pk))
                writer.commit()
            finally:
                lock.release()

    def get_index(self, search_model):
        storage = self.get_storage()

        schema = self.get_search_model_schema(search_model=search_model)

        try:
            index = storage.open_index(
                indexname=search_model.get_full_name(), schema=schema
            )
        except EmptyIndexError:
            index = storage.create_index(
                indexname=search_model.get_full_name(), schema=schema
            )

        return index

    def get_resolved_field_map(self, search_model):
        if search_model not in self._resolved_field_maps:

            result = {}
            for search_field in self.get_search_model_fields(search_model=search_model):
                whoosh_field_type = DJANGO_TO_WHOOSH_FIELD_MAP.get(
                    search_field.get_model_field().__class__
                )
                if whoosh_field_type:
                    result[search_field.get_full_name()] = whoosh_field_type
                else:
                    logger.warning(
                        'unknown field type "%s" for model "%s"',
                        search_field.get_full_name(),
                        search_model.get_full_name()
                    )

            self._resolved_field_maps[search_model] = result

        return self._resolved_field_maps[search_model]

    def get_search_model_fields(self, search_model):
        result = search_model.search_fields.copy()
        result.append(
            SearchField(search_model=search_model, field='id', label='ID')
        )
        return result

    def get_search_model_schema(self, search_model):
        field_map = self.get_resolved_field_map(search_model=search_model)
        schema_kwargs = {key: value['field'] for key, value in field_map.items()}
        return whoosh.fields.Schema(**schema_kwargs)

    def get_storage(self):
        return FileStorage(path=self.index_path)

    def index_instance(self, instance, exclude_set=None):
        try:
            lock = LockingBackend.get_backend().acquire_lock(
                name='dynamic_search_whoosh_index_instance'
            )
        except LockError:
            raise
        else:
            try:
                # Use a private method to allow using a single lock for
                # all recursions.
                self._index_instance(
                    instance=instance, exclude_set=exclude_set
                )
            finally:
                lock.release()

    def _index_instance(self, instance, exclude_set=None):
        if not exclude_set:
            exclude_set = set()

        # Avoid infinite recursion.
        if instance in exclude_set:
            return

        exclude_set.add(instance)

        try:
            search_model = SearchModel.get_for_model(instance=instance)
        except KeyError:
            """
            A KeyError is not fatal. It means search is not configured
            for this instance but we still need to check if one of its
            field's related models are configure for search and need
            to be updated.
            """
        else:
            index = self.get_index(search_model=search_model)

            writer = index.writer()
            kwargs = search_model.sieve(
                field_map=self.get_resolved_field_map(search_model=search_model), instance=instance
            )
            writer.delete_by_term('id', str(instance.pk))
            try:
                writer.add_document(**kwargs)
                writer.commit()
            except Exception as exception:
                logger.error(
                    'Unexpected exception while indexing object id: %s, '
                    'search model: %s, index data: %s, raw data: %s, '
                    'field map: %s; %s', search_model.get_full_name(),
                    instance.pk, kwargs, instance.__dict__,
                    self.get_resolved_field_map(search_model=search_model),
                    exception, exc_info=True
                )
                raise

        for field_class in instance._meta.get_fields():
            # Only to recursive indexing for related models that are
            # known to have a search configuration.
            if field_class.related_model and field_class.related_model in SearchModel._model_search_relationships.get(instance._meta.model, ()):
                field_instance = getattr(instance, field_class.name, None)

                if field_instance:
                    try:
                        # Try as a many field.
                        results = field_instance.all()
                    except AttributeError:
                        # Try as a one to one field.
                        try:
                            results = [field_instance.get()]
                        except AttributeError:
                            # It is neither then it must be a
                            # foreign key.
                            results = [field_instance]

                    for instance in results:
                        self._index_instance(
                            instance=instance, exclude_set=exclude_set
                        )

    def index_search_model(self, search_model):
        index = self.get_index(search_model=search_model)

        # Clear the model index
        self.get_storage().create_index(
            index.schema, indexname=search_model.get_full_name()
        )

        for instance in search_model.model._meta.default_manager.all():
            self.index_instance(instance=instance)



from django import forms
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.classes import ModelField, ModelFieldRelated
from mayan.apps.documents.models import Document
from mayan.apps.templating.fields import ModelTemplateField

from .models import SmartLink, SmartLinkCondition


class SmartLinkForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['dynamic_label'] = ModelTemplateField(
            initial_help_text=self.fields['dynamic_label'].help_text,
            label=self.fields['dynamic_label'].label, model=Document,
            model_variable='document', required=False
        )

    class Meta:
        fields = ('label', 'dynamic_label', 'enabled')
        model = SmartLink


class SmartLinkConditionForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        choices = []
        choices.append(
            (
                ModelField.class_label, ModelField.get_choices_for(
                    model=Document,
                )
            )
        )
        choices.append(
            (
                ModelFieldRelated.class_label, ModelFieldRelated.get_choices_for(
                    model=Document,
                )
            )
        )

        self.fields['foreign_document_data'] = forms.ChoiceField(
            choices=choices, label=_('Foreign document field')
        )
        self.fields['expression'] = ModelTemplateField(
            initial_help_text=self.fields['expression'].help_text,
            label=self.fields['expression'].label, model=Document,
            model_variable='document', required=False
        )

    class Meta:
        model = SmartLinkCondition
        exclude = ('smart_link',)



from mayan.apps.testing.tests.base import BaseTestCase

from ..models import ErrorLogPartitionEntry

from .mixins import LoggingTextMixin


class LoggingModelTestCase(LoggingTextMixin, BaseTestCase):
    def setUp(self):
        super().setUp()
        self._create_error_log_test_object()
        self._create_error_log_entry()

    def test_entries_limit(self):
        self.error_log.limit = 3
        self.test_object.error_log.create(text='1')
        self.test_object.error_log.create(text='2')
        self.test_object.error_log.create(text='3')
        self.test_object.error_log.create(text='4')

        self.assertEqual(
            list(self.test_object.error_log.values_list('text', flat=True)),
            ['2', '3', '4']
        )

    def test_object_deletion(self):
        self.test_object.delete()
        self.assertEqual(ErrorLogPartitionEntry.objects.count(), 0)



from django.contrib.auth.models import Group
from django.template import RequestContext
from django.urls import reverse_lazy
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from mayan.apps.user_management.permissions import permission_group_edit
from mayan.apps.views.generics import (
    AddRemoveView, SingleObjectCreateView, SingleObjectDeleteView,
    SingleObjectEditView, SingleObjectListView
)

from .icons import icon_role_list
from .links import link_role_create
from .models import Role, StoredPermission
from .permissions import (
    permission_role_view, permission_role_create, permission_role_delete,
    permission_role_edit
)


class GroupRolesView(AddRemoveView):
    main_object_method_add_name = 'roles_add'
    main_object_method_remove_name = 'roles_remove'
    main_object_model = Group
    main_object_permission = permission_group_edit
    main_object_pk_url_kwarg = 'group_id'
    secondary_object_model = Role
    secondary_object_permission = permission_role_edit
    list_available_title = _('Available roles')
    list_added_title = _('Group roles')
    related_field = 'roles'

    def get_actions_extra_kwargs(self):
        return {'_event_actor': self.request.user}

    def get_extra_context(self):
        return {
            'object': self.main_object,
            'title': _('Roles of group: %s') % self.main_object,
        }


class RoleCreateView(SingleObjectCreateView):
    fields = ('label',)
    model = Role
    view_permission = permission_role_create
    post_action_redirect = reverse_lazy(viewname='permissions:role_list')

    def get_extra_context(self):
        return {'title': _('Create new role')}

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class RoleDeleteView(SingleObjectDeleteView):
    model = Role
    object_permission = permission_role_delete
    pk_url_kwarg = 'role_id'
    post_action_redirect = reverse_lazy(viewname='permissions:role_list')


class RoleEditView(SingleObjectEditView):
    fields = ('label',)
    model = Role
    object_permission = permission_role_edit
    pk_url_kwarg = 'role_id'

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class SetupRoleMembersView(AddRemoveView):
    main_object_method_add_name = 'groups_add'
    main_object_method_remove_name = 'groups_remove'
    main_object_model = Role
    main_object_permission = permission_role_edit
    main_object_pk_url_kwarg = 'role_id'
    secondary_object_model = Group
    secondary_object_permission = permission_group_edit
    list_available_title = _('Available groups')
    list_added_title = _('Role groups')
    related_field = 'groups'

    def get_actions_extra_kwargs(self):
        return {'_event_actor': self.request.user}

    def get_extra_context(self):
        return {
            'object': self.main_object,
            'title': _('Groups of role: %s') % self.main_object,
            'subtitle': _(
                'Add groups to be part of a role. They will '
                'inherit the role\'s permissions and access controls.'
            ),
        }


class SetupRolePermissionsView(AddRemoveView):
    grouped = True
    main_object_method_add_name = 'permissions_add'
    main_object_method_remove_name = 'permissions_remove'
    main_object_model = Role
    main_object_permission = permission_role_edit
    main_object_pk_url_kwarg = 'role_id'
    list_available_title = _('Available permissions')
    list_added_title = _('Granted permissions')
    related_field = 'permissions'
    secondary_object_model = StoredPermission

    def generate_choices(self, queryset):
        namespaces_dictionary = {}

        # Sort permissions by their translatable label
        object_list = sorted(
            queryset,
            key=lambda permission: permission.volatile_permission.label
        )

        # Group permissions by namespace
        for permission in object_list:
            namespaces_dictionary.setdefault(
                permission.volatile_permission.namespace.label,
                []
            )
            namespaces_dictionary[
                permission.volatile_permission.namespace.label
            ].append(
                (permission.pk, force_text(s=permission))
            )

        # Sort permissions by their translatable namespace label
        return sorted(namespaces_dictionary.items())

    def get_actions_extra_kwargs(self):
        return {'_event_actor': self.request.user}

    def get_extra_context(self):
        return {
            'object': self.main_object,
            'subtitle': _(
                'Permissions granted here will apply to the entire system '
                'and all objects.'
            ),
            'title': _('Permissions for role: %s') % self.main_object,
        }


class RoleListView(SingleObjectListView):
    model = Role
    object_permission = permission_role_view

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_role_list,
            'no_results_main_link': link_role_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Roles are authorization units. They contain '
                'user groups which inherit the role permissions for the '
                'entire system. Roles can also part of access '
                'controls lists. Access controls list are permissions '
                'granted to a role for specific objects which its group '
                'members inherit.'
            ),
            'no_results_title': _('There are no roles'),
            'title': _('Roles'),
        }



from django.apps import apps


def handler_user_theme_setting_create(sender, instance, created, **kwargs):
    UserThemeSetting = apps.get_model(
        app_label='appearance', model_name='UserThemeSetting'
    )

    if created:
        UserThemeSetting.objects.create(user=instance)



# -*- coding: utf-8 -*-

DOCUMENT_METADATA_ADD_ACTION_CLASS_PATH = 'mayan.apps.metadata.workflow_actions.DocumentMetadataAddAction'
DOCUMENT_METADATA_EDIT_ACTION_CLASS_PATH = 'mayan.apps.metadata.workflow_actions.DocumentMetadataEditAction'
DOCUMENT_METADATA_REMOVE_ACTION_CLASS_PATH = 'mayan.apps.metadata.workflow_actions.DocumentMetadataRemoveAction'
TEST_CORRECT_LOOKUP_VALUE = '1'
TEST_DATE_PARSER = 'mayan.apps.metadata.parsers.DateParser'
TEST_DATE_VALIDATOR = 'mayan.apps.metadata.validators.DateValidator'
TEST_DEFAULT_VALUE = 'test'
TEST_INCORRECT_LOOKUP_VALUE = '0'
TEST_INVALID_DATE = '___________'
TEST_LOOKUP_TEMPLATE = '1,2,3'
TEST_METADATA_TYPE_DEFAULT_VALUE = 'default value'
TEST_METADATA_TYPE_LABEL = 'test metadata type label'
TEST_METADATA_TYPE_NAME = 'test_metadata_type_name'
TEST_METADATA_TYPE_LABEL_EDITED = 'test metadata type label edited'
TEST_METADATA_TYPE_NAME_EDITED = 'test_metadata_type_name_edited'
TEST_METADATA_VALUE = 'test value'
TEST_METADATA_VALUE_EDITED = 'test value edited'
TEST_METADATA_VALUE_UNICODE = 'español'
TEST_METADATA_VALUE_WITH_AMPERSAND = 'first value & second value'
TEST_PARSED_VALID_DATE = '2001-01-01'
TEST_VALID_DATE = '2001-1-1'



from rest_framework import serializers


class AJAXTemplateSerializer(serializers.Serializer):
    hex_hash = serializers.CharField(read_only=True)
    name = serializers.CharField(read_only=True)
    html = serializers.CharField(read_only=True)
    url = serializers.HyperlinkedIdentityField(
        lookup_field='name', lookup_url_kwarg='name',
        view_name='rest_api:template-detail'
    )



import logging

from mayan.apps.document_indexing.tasks import task_index_document

logger = logging.getLogger(name=__name__)


def handler_index_document(sender, **kwargs):
    if kwargs['action'] in ('post_add', 'post_remove'):
        for pk in kwargs['pk_set']:
            task_index_document.apply_async(kwargs={'document_id': pk})


def handler_tag_pre_delete(sender, **kwargs):
    for document in kwargs['instance'].documents.all():
        # Remove each of the documents from the tag
        # Trigger the m2m_changed signal for each document so they can be
        # reindexed
        kwargs['instance'].documents.remove(document)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_('Authentication'), name='authentication'
)

event_user_impersonation_ended = namespace.add_event_type(
    label=_('User impersonation ended'), name='user_impersonation_ended'
)
event_user_impersonation_started = namespace.add_event_type(
    label=_('User impersonation started'), name='user_impersonation_started'
)
event_user_logged_in = namespace.add_event_type(
    label=_('User logged in'), name='user_logged_in'
)
event_user_logged_out = namespace.add_event_type(
    label=_('User logged out'), name='user_logged_out'
)



from django.db import migrations, models
import django.db.models.deletion


def operation_initialize_file_metadata_settings(apps, schema_editor):
    DocumentType = apps.get_model(
        app_label='documents', model_name='DocumentType'
    )
    DocumentTypeSettings = apps.get_model(
        app_label='file_metadata', model_name='DocumentTypeSettings'
    )

    for document_type in DocumentType.objects.using(alias=schema_editor.connection.alias).all():
        DocumentTypeSettings.objects.using(
            alias=schema_editor.connection.alias
        ).create(document_type=document_type)


def operation_initialize_file_metadata_settings_reverse(apps, schema_editor):
    DocumentTypeSettings = apps.get_model(
        app_label='file_metadata', model_name='DocumentTypeSettings'
    )
    DocumentTypeSettings.objects.using(alias=schema_editor.connection.alias).delete()


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ('documents', '0047_auto_20180917_0737'),
    ]

    operations = [
        migrations.CreateModel(
            name='DocumentTypeSettings',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'auto_process', models.BooleanField(
                        default=True, verbose_name='Automatically queue '
                        'newly created documents for processing.'
                    )
                ),
                (
                    'document_type', models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='file_metadata_settings',
                        to='documents.DocumentType',
                        verbose_name='Document type'
                    )
                ),
            ],
            options={
                'verbose_name': 'Document type settings',
                'verbose_name_plural': 'Document types settings',
            },
        ),
        migrations.CreateModel(
            name='DocumentVersionDriverEntry',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'document_version', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='file_metadata_drivers',
                        to='documents.DocumentVersion',
                        verbose_name='Document version'
                    )
                ),
            ],
            options={
                'ordering': ('document_version', 'driver'),
                'verbose_name': 'Document version driver entry',
                'verbose_name_plural': 'Document version driver entries',
            },
        ),
        migrations.CreateModel(
            name='FileMetadataEntry',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'key', models.CharField(
                        db_index=True, help_text='Name of the file metadata '
                        'entry.', max_length=255, verbose_name='Key'
                    )
                ),
                (
                    'value', models.CharField(
                        db_index=True, help_text='Value of the file metadata '
                        'entry.', max_length=255, verbose_name='Value'
                    )
                ),
                (
                    'document_version_driver_entry', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='entries',
                        to='file_metadata.DocumentVersionDriverEntry',
                        verbose_name='Document version driver entry'
                    )
                ),
            ],
            options={
                'ordering': ('key', 'value'),
                'verbose_name': 'File metadata entry',
                'verbose_name_plural': 'File metadata entries',
            },
        ),
        migrations.CreateModel(
            name='StoredDriver',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'driver_path', models.CharField(
                        max_length=255, verbose_name='Driver path'
                    )
                ),
                (
                    'internal_name', models.CharField(
                        db_index=True, max_length=128,
                        verbose_name='Internal name'
                    )
                ),
            ],
            options={
                'ordering': ('internal_name',),
                'verbose_name': 'Driver',
                'verbose_name_plural': 'Drivers',
            },
        ),
        migrations.AddField(
            model_name='documentversiondriverentry',
            name='driver',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='driver_entries',
                to='file_metadata.StoredDriver', verbose_name='Driver'
            ),
        ),
        migrations.AlterUniqueTogether(
            name='documentversiondriverentry',
            unique_together=set([('driver', 'document_version')]),
        ),
        migrations.RunPython(
            code=operation_initialize_file_metadata_settings,
            reverse_code=operation_initialize_file_metadata_settings_reverse
        )
    ]



class BaseCommonException(Exception):
    """
    Base exception for the common app
    """


class ResolverError(Exception):
    """
    The resolver class was not able to resolve the requested attribute.
    This is a not fatal exception and just makes the resolver pipeline
    try the next resolver class in the list.
    """


class ResolverPipelineError(Exception):
    """
    Raised when the resolver pipeline exhausted the list of resolvers
    and nothing new was returned. This means that the requested
    attribute does not exists.
    """


class NonUniqueError(BaseCommonException):
    """
    Raised when attempting to add duplicate values to a type that expects
    them to be unique.
    """



from mayan.apps.document_indexing.models import (
    IndexInstanceNode, IndexTemplate
)
from mayan.apps.document_indexing.tests.literals import TEST_INDEX_TEMPLATE_LABEL
from mayan.apps.documents.tests.base import DocumentTestMixin
from mayan.apps.testing.tests.base import BaseTransactionTestCase

from .literals import (
    TEST_FILE_METADATA_INDEX_NODE_TEMPLATE, TEST_FILE_METADATA_VALUE
)


class IndexingTestCase(DocumentTestMixin, BaseTransactionTestCase):
    auto_upload_test_document = False

    def test_indexing(self):
        self.test_index_template = IndexTemplate.objects.create(
            label=TEST_INDEX_TEMPLATE_LABEL
        )
        self.test_index_template.document_types.add(self.test_document_type)

        root = self.test_index_template.template_root
        self.test_index_template.node_templates.create(
            parent=root, expression=TEST_FILE_METADATA_INDEX_NODE_TEMPLATE,
            link_documents=True
        )

        self._upload_test_document()

        self.test_document.submit_for_file_metadata_processing()

        self.assertTrue(
            self.test_document in IndexInstanceNode.objects.get(
                value=TEST_FILE_METADATA_VALUE
            ).documents.all()
        )



import logging

from furl import furl

from django.contrib import messages
from django.contrib.contenttypes.models import ContentType
from django.urls import reverse
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _, ungettext
from django.views.generic import RedirectView

from mayan.apps.common.classes import ModelQueryFields
from mayan.apps.common.settings import setting_home_view
from mayan.apps.converter.literals import DEFAULT_ROTATION, DEFAULT_ZOOM_LEVEL
from mayan.apps.views.generics import (
    FormView, MultipleObjectConfirmActionView, SingleObjectDeleteView,
    SingleObjectListView, SimpleView
)
from mayan.apps.views.mixins import ExternalObjectViewMixin
from mayan.apps.views.utils import resolve

from ..forms.document_version_page_forms import (
    DocumentVersionPageForm, DocumentVersionPageMappingFormSet
)
from ..icons import (
    icon_document_version_page_list, icon_document_version_page_list_remap
)
from ..links.document_version_page_links import (
    link_document_version_page_list_remap,
    link_document_version_page_list_reset
)
from ..models.document_version_models import DocumentVersion
from ..models.document_version_page_models import DocumentVersionPage
from ..permissions import (
    permission_document_version_edit, permission_document_version_view
)
from ..settings import (
    setting_rotation_step, setting_zoom_percent_step, setting_zoom_max_level,
    setting_zoom_min_level
)
from ..tasks import task_document_version_page_list_reset

__all__ = (
    'DocumentVersionPageListView',
    'DocumentVersionPageNavigationFirst', 'DocumentVersionPageNavigationLast',
    'DocumentVersionPageNavigationNext', 'DocumentVersionPageNavigationPrevious',
    'DocumentVersionPageView', 'DocumentVersionPageViewResetView',
    'DocumentVersionPageInteractiveTransformation', 'DocumentVersionPageZoomInView',
    'DocumentVersionPageZoomOutView', 'DocumentVersionPageRotateLeftView',
    'DocumentVersionPageRotateRightView'
)
logger = logging.getLogger(name=__name__)


class DocumentVersionPageDeleteView(SingleObjectDeleteView):
    object_permission = permission_document_version_edit
    pk_url_kwarg = 'document_version_page_id'
    source_queryset = DocumentVersionPage.valid

    def get_extra_context(self):
        return {
            'message': _(
                'The page number of this page will be skipped. If you '
                'want to achieve sequential page numbering, use the '
                'page remap action instead.'
            ),
            'object': self.object,
            'title': _('Delete document version page %s ?') % self.object,
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_post_action_redirect(self):
        return reverse(
            viewname='documents:document_version_page_list', kwargs={
                'document_version_id': self.object.document_version_id
            }
        )


class DocumentVersionPageListView(ExternalObjectViewMixin, SingleObjectListView):
    external_object_permission = permission_document_version_view
    external_object_pk_url_kwarg = 'document_version_id'
    external_object_queryset = DocumentVersion.valid

    def get_extra_context(self):
        return {
            'hide_object': True,
            'list_as_items': True,
            'no_results_icon': icon_document_version_page_list,
            'no_results_main_link': link_document_version_page_list_reset.resolve(
                request=self.request, resolved_object=self.external_object
            ),
            'no_results_secondary_links': [
                link_document_version_page_list_remap.resolve(
                    request=self.request, resolved_object=self.external_object
                ),
            ],
            'no_results_text': _(
                'Document version pages are links to actual content pages. '
                'Create them using the page remap actions or the page '
                'reset action.'
            ),
            'no_results_title': _('No document version pages available'),
            'object': self.external_object,
            'title': _('Pages of document version: %s') % self.external_object,
        }

    def get_source_queryset(self):
        queryset = ModelQueryFields.get(model=DocumentVersionPage).get_queryset()
        return queryset.filter(pk__in=self.external_object.pages.all())


class DocumentVersionPageListRemapView(ExternalObjectViewMixin, FormView):
    external_object_permission = permission_document_version_edit
    external_object_pk_url_kwarg = 'document_version_id'
    external_object_queryset = DocumentVersion.valid
    form_class = DocumentVersionPageMappingFormSet

    def form_valid(self, form):
        annotated_content_object_list = []

        for row in form.forms:
            page_number = int(row.cleaned_data['target_page_number'])
            if page_number:
                content_type = ContentType.objects.get(
                    pk=row.cleaned_data['source_content_type']
                )
                content_object = content_type.get_object_for_this_type(
                    pk=row.cleaned_data['source_object_id']
                )

                annotated_content_object_list.append(
                    {
                        'page_number': page_number,
                        'content_object': content_object
                    }
                )

        self.external_object.pages_remap(
            annotated_content_object_list=annotated_content_object_list
        )
        return super().form_valid(form=form)

    def get_form_extra_kwargs(self):
        target_page_number_choices = [(0, _('None'))]

        page_index = 1

        for document_file in self.external_object.document.files.all():
            for document_file_page in document_file.pages.all():
                target_page_number_choices.append(
                    (page_index, page_index)
                )

                page_index = page_index + 1

        return {
            'form_extra_kwargs': {
                'target_page_number_choices': target_page_number_choices
            }
        }

    def get_extra_context(self):
        return {
            'form_display_mode_table': True,
            'hide_object': True,
            'list_as_items': True,
            'no_results_icon': icon_document_version_page_list_remap,
            'no_results_text': _(
                'There are no sources available to remap for this document '
                'version.'
            ),
            'no_results_title': _('No page sources available'),
            'object': self.external_object,
            'title': _(
                'Remap pages of document version: %s'
            ) % self.external_object,
        }

    def get_initial(self):
        initial = []

        content_object_dictionary_list = self.external_object.get_source_content_object_dictionary_list()

        for content_object_dictionary in content_object_dictionary_list:
            content_object = content_object_dictionary['content_type'].get_object_for_this_type(
                id=content_object_dictionary['object_id']
            )

            # The same source object could have been assigned to multiple
            # document version pages.
            document_version_pages = self.external_object.pages.filter(
                content_type=content_object_dictionary['content_type'],
                object_id=content_object_dictionary['object_id']
            )

            if document_version_pages:
                document_version_page_page_number = document_version_pages.first().page_number
            else:
                document_version_page_page_number = 0

            row = {
                'source_content_type': content_object_dictionary['content_type'].pk,
                'source_object_id': content_object_dictionary['object_id'],
                'source_thumbnail': content_object,
                'source_label': '{}: {}'.format(content_object_dictionary['content_type'].name, content_object),
                'target_page_number': document_version_page_page_number
            }

            initial.append(row)

        return initial

    def get_post_action_redirect(self):
        return reverse(
            viewname='documents:document_version_page_list', kwargs={
                'document_version_id': self.external_object.pk
            }
        )


class DocumentVersionPageListResetView(MultipleObjectConfirmActionView):
    object_permission = permission_document_version_edit
    pk_url_kwarg = 'document_version_id'
    source_queryset = DocumentVersion.valid
    success_message = _(
        '%(count)d document version queued for page list reset.'
    )
    success_message_plural = _(
        '%(count)d document versions queued for page list reset.'
    )

    def get_extra_context(self):
        queryset = self.object_list

        result = {
            'message': _(
                'The page list will match that of the latest document file.'
            ),
            'title': ungettext(
                singular='Reset the page list of the selected document version?',
                plural='Reset the page list of the selected document versions?',
                number=queryset.count()
            )
        }

        if queryset.count() == 1:
            result.update(
                {
                    'object': queryset.first(),
                    'title': _(
                        'Reset the page list of document version: %s?'
                    ) % queryset.first()
                }
            )

        return result

    def object_action(self, form, instance):
        task_document_version_page_list_reset.apply_async(
            kwargs={'document_version_id': instance.pk}
        )


class DocumentVersionPageNavigationBase(ExternalObjectViewMixin, RedirectView):
    external_object_permission = permission_document_version_view
    external_object_pk_url_kwarg = 'document_version_page_id'
    external_object_queryset = DocumentVersionPage.valid

    def get_redirect_url(self, *args, **kwargs):
        """
        Attempt to jump to the same kind of view but resolved to a new
        object of the same kind.
        """
        previous_url = self.request.META.get('HTTP_REFERER', None)

        if not previous_url:
            try:
                previous_url = self.external_object.get_absolute_url()
            except AttributeError:
                previous_url = reverse(viewname=setting_home_view.value)

        parsed_url = furl(url=previous_url)

        # Obtain the view name to be able to resolve it back with new keyword
        # arguments.
        resolver_match = resolve(path=force_text(s=parsed_url.path))

        new_kwargs = self.get_new_kwargs()

        if set(new_kwargs) == set(resolver_match.kwargs):
            # It is the same type of object, reuse the URL to stay in the
            # same kind of view but pointing to a new object
            url = reverse(
                viewname=resolver_match.view_name, kwargs=new_kwargs
            )
        else:
            url = parsed_url.path

        # Update just the path to retain the querystring in case there is
        # transformation data.
        parsed_url.path = url

        return parsed_url.tostr()


class DocumentVersionPageNavigationFirst(DocumentVersionPageNavigationBase):
    def get_new_kwargs(self):
        return {
            'document_version_page_id': self.external_object.siblings.first().pk
        }


class DocumentVersionPageNavigationLast(DocumentVersionPageNavigationBase):
    def get_new_kwargs(self):
        return {
            'document_version_page_id': self.external_object.siblings.last().pk
        }


class DocumentVersionPageNavigationNext(DocumentVersionPageNavigationBase):
    def get_new_kwargs(self):
        new_document_version_page = self.external_object.siblings.filter(
            page_number__gt=self.external_object.page_number
        ).first()
        if new_document_version_page:
            return {'document_version_page_id': new_document_version_page.pk}
        else:
            messages.warning(
                message=_(
                    'There are no more pages in this document'
                ), request=self.request
            )
            return {'document_version_page_id': self.external_object.pk}


class DocumentVersionPageNavigationPrevious(DocumentVersionPageNavigationBase):
    def get_new_kwargs(self):
        new_document_version_page = self.external_object.siblings.filter(
            page_number__lt=self.external_object.page_number
        ).last()
        if new_document_version_page:
            return {'document_version_page_id': new_document_version_page.pk}
        else:
            messages.warning(
                message=_(
                    'You are already at the first page of this document'
                ), request=self.request
            )
            return {'document_version_page_id': self.external_object.pk}


class DocumentVersionPageView(ExternalObjectViewMixin, SimpleView):
    external_object_permission = permission_document_version_view
    external_object_pk_url_kwarg = 'document_version_page_id'
    external_object_queryset = DocumentVersionPage.valid
    template_name = 'appearance/generic_form.html'

    def get_extra_context(self):
        zoom = int(self.request.GET.get('zoom', DEFAULT_ZOOM_LEVEL))
        rotation = int(self.request.GET.get('rotation', DEFAULT_ROTATION))

        document_version_page_form = DocumentVersionPageForm(
            instance=self.external_object, rotation=rotation, zoom=zoom
        )

        base_title = _('Image of: %s') % self.external_object

        if zoom != DEFAULT_ZOOM_LEVEL:
            zoom_text = '({}%)'.format(zoom)
        else:
            zoom_text = ''

        return {
            'form': document_version_page_form,
            'hide_labels': True,
            'object': self.external_object,
            'rotation': rotation,
            'title': ' '.join((base_title, zoom_text)),
            'read_only': True,
            'zoom': zoom,
        }


class DocumentVersionPageViewResetView(RedirectView):
    pattern_name = 'documents:document_version_page_view'


class DocumentVersionPageInteractiveTransformation(
    ExternalObjectViewMixin, RedirectView
):
    external_object_permission = permission_document_version_view
    external_object_pk_url_kwarg = 'document_version_page_id'
    external_object_queryset = DocumentVersionPage.valid

    def get_object(self):
        return self.external_object

    def get_redirect_url(self, *args, **kwargs):
        query_dict = {
            'rotation': self.request.GET.get('rotation', DEFAULT_ROTATION),
            'zoom': self.request.GET.get('zoom', DEFAULT_ZOOM_LEVEL)
        }

        url = furl(
            args=query_dict, path=reverse(
                viewname='documents:document_version_page_view', kwargs={
                    'document_version_page_id': self.external_object.pk
                }
            )

        )

        self.transformation_function(query_dict=query_dict)
        # Refresh query_dict to args reference
        url.args = query_dict

        return url.tostr()


class DocumentVersionPageZoomInView(
    DocumentVersionPageInteractiveTransformation
):
    def transformation_function(self, query_dict):
        zoom = int(query_dict['zoom']) + setting_zoom_percent_step.value

        if zoom > setting_zoom_max_level.value:
            zoom = setting_zoom_max_level.value

        query_dict['zoom'] = zoom


class DocumentVersionPageZoomOutView(
    DocumentVersionPageInteractiveTransformation
):
    def transformation_function(self, query_dict):
        zoom = int(query_dict['zoom']) - setting_zoom_percent_step.value

        if zoom < setting_zoom_min_level.value:
            zoom = setting_zoom_min_level.value

        query_dict['zoom'] = zoom


class DocumentVersionPageRotateLeftView(
    DocumentVersionPageInteractiveTransformation
):
    def transformation_function(self, query_dict):
        query_dict['rotation'] = (
            int(query_dict['rotation']) - setting_rotation_step.value
        ) % 360


class DocumentVersionPageRotateRightView(
    DocumentVersionPageInteractiveTransformation
):
    def transformation_function(self, query_dict):
        query_dict['rotation'] = (
            int(query_dict['rotation']) + setting_rotation_step.value
        ) % 360



from django.utils.translation import ugettext_lazy as _

from mayan.apps.permissions import PermissionNamespace

namespace = PermissionNamespace(
    label=_('Authentication'), name='authentication'
)

permission_users_impersonate = namespace.add_permission(
    label=_('Impersonate users'), name='users_impersonate'
)



import logging
import subprocess

from django.db import models
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _

from mayan.apps.storage.utils import TemporaryFile

from ..classes import PseudoFile, SourceUploadedFile
from ..exceptions import SourceException
from ..literals import (
    SCANNER_ADF_MODE_CHOICES, SCANNER_MODE_CHOICES, SCANNER_MODE_COLOR,
    SCANNER_SOURCE_CHOICES, SOURCE_CHOICE_SANE_SCANNER,
)
from ..settings import setting_scanimage_path

from .base import InteractiveSource

__all__ = ('SaneScanner',)
logger = logging.getLogger(name=__name__)


class SaneScanner(InteractiveSource):
    can_compress = False
    is_interactive = True
    source_type = SOURCE_CHOICE_SANE_SCANNER

    device_name = models.CharField(
        max_length=255,
        help_text=_('Device name as returned by the SANE backend.'),
        verbose_name=_('Device name')
    )
    mode = models.CharField(
        blank=True, choices=SCANNER_MODE_CHOICES, default=SCANNER_MODE_COLOR,
        help_text=_(
            'Selects the scan mode (e.g., lineart, monochrome, or color). '
            'If this option is not supported by your scanner, leave it blank.'
        ), max_length=16, verbose_name=_('Mode')
    )
    resolution = models.PositiveIntegerField(
        blank=True, null=True, help_text=_(
            'Sets the resolution of the scanned image in DPI (dots per inch). '
            'Typical value is 200. If this option is not supported by your '
            'scanner, leave it blank.'
        ), verbose_name=_('Resolution')
    )
    source = models.CharField(
        blank=True, choices=SCANNER_SOURCE_CHOICES, help_text=_(
            'Selects the scan source (such as a document-feeder). If this '
            'option is not supported by your scanner, leave it blank.'
        ), max_length=32, null=True, verbose_name=_('Paper source')
    )
    adf_mode = models.CharField(
        blank=True, choices=SCANNER_ADF_MODE_CHOICES,
        help_text=_(
            'Selects the document feeder mode (simplex/duplex). If this '
            'option is not supported by your scanner, leave it blank.'
        ), max_length=16, verbose_name=_('ADF mode')
    )

    objects = models.Manager()

    class Meta:
        verbose_name = _('SANE Scanner')
        verbose_name_plural = _('SANE Scanners')

    def clean_up_upload_file(self, upload_file_object):
        pass

    def execute_command(self, arguments):
        command_line = [
            setting_scanimage_path.value
        ]
        command_line.extend(arguments)

        with TemporaryFile() as stderr_file_object:
            stdout_file_object = TemporaryFile()

            try:
                logger.debug('Scan command line: %s', command_line)
                subprocess.check_call(
                    command_line, stdout=stdout_file_object,
                    stderr=stderr_file_object
                )
            except subprocess.CalledProcessError:
                stderr_file_object.seek(0)
                error_message = stderr_file_object.read()
                logger.error(
                    'Exception while executing scanning command for source:%s ; %s', self,
                    error_message
                )

                message = _(
                    'Error while executing scanning command '
                    '"%(command_line)s"; %(error_message)s'
                ) % {
                    'command_line': ' '.join(command_line),
                    'error_message': error_message
                }
                self.error_log.create(text=message)
                raise SourceException(message)
            else:
                stdout_file_object.seek(0)
                self.error_log.all().delete()
                return stdout_file_object

    def get_upload_file_object(self, form_data):
        arguments = [
            '-d', self.device_name, '--format', 'tiff',
        ]

        if self.resolution:
            arguments.extend(
                ['--resolution', '{}'.format(self.resolution)]
            )

        if self.mode:
            arguments.extend(
                ['--mode', self.mode]
            )

        if self.source:
            arguments.extend(
                ['--source', self.source]
            )

        if self.adf_mode:
            arguments.extend(
                ['--adf-mode', self.adf_mode]
            )

        file_object = self.execute_command(arguments=arguments)

        return SourceUploadedFile(
            source=self, file=PseudoFile(
                file=file_object, name='scan {}'.format(now())
            )
        )



TEST_INDEX_TEMPLATE_LABEL = 'test label'
TEST_INDEX_TEMPLATE_LABEL_EDITED = 'test edited label'
TEST_INDEX_TEMPLATE_SLUG = 'test_slug'
TEST_INDEX_TEMPLATE_NODE_EXPRESSION = '{{ document.label }}'
TEST_INDEX_TEMPLATE_NODE_EXPRESSION_EDITED = '{{ document.description }}'
TEST_METADATA_TYPE_LABEL = 'test metadata label'
TEST_METADATA_TYPE_NAME = 'test_metadata_name'
TEST_INDEX_TEMPLATE_METADATA_EXPRESSION = '{{ document.metadata_value_of.%s }}' % TEST_METADATA_TYPE_NAME
TEST_INDEX_TEMPLATE_DOCUMENT_LABEL_EXPRESSION = '{{ document.label }}'
TEST_INDEX_TEMPLATE_DOCUMENT_DESCRIPTION_EXPRESSION = '{{ document.description }}'



from mayan.apps.documents.tests.base import GenericDocumentTestCase

from ..models import DuplicateBackendEntry, StoredDuplicateBackend

from .mixins import DuplicatedDocumentTestMixin


class DuplicatedDocumentModelTestCase(
    DuplicatedDocumentTestMixin, GenericDocumentTestCase
):
    def test_duplicates_after_delete(self):
        self._upload_duplicate_document()
        self.test_documents[1].delete()
        self.test_documents[1].delete()

        self.assertEqual(
            DuplicateBackendEntry.objects.filter(
                document=self.test_documents[0]
            ).count(), 0
        )

    def test_duplicates_after_trash(self):
        self._upload_duplicate_document()
        self.test_documents[1].delete()

        self.assertFalse(
            self.test_documents[1] in DuplicateBackendEntry.objects.get_duplicates_of(
                document=self.test_documents[0]
            )
        )

    def test_duplicate_scan_after_upload(self):
        self._upload_duplicate_document()

        self.assertTrue(
            self.test_documents[1] in DuplicateBackendEntry.objects.get_duplicates_of(
                document=self.test_documents[0]
            )
        )
        self.assertTrue(
            self.test_documents[0] in DuplicateBackendEntry.objects.get_duplicates_of(
                document=self.test_documents[1]
            )
        )

    def test_double_duplicated_entry_for_a_single_document(self):
        self._upload_duplicate_document()

        # Should not return an error.
        StoredDuplicateBackend.objects.scan_document(
            document=self.test_documents[0]
        )



import logging

from django.conf import settings
from django.db import models
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from ..managers import RecentlyAccessedDocumentManager

from .document_models import Document

__all__ = ('RecentlyAccessedDocument',)
logger = logging.getLogger(name=__name__)


class RecentlyAccessedDocument(models.Model):
    """
    Keeps a list of the n most recent accessed or created document for
    a given user
    """
    user = models.ForeignKey(
        db_index=True, editable=False, on_delete=models.CASCADE,
        to=settings.AUTH_USER_MODEL, verbose_name=_('User')
    )
    document = models.ForeignKey(
        editable=False, on_delete=models.CASCADE, related_name='recent',
        to=Document, verbose_name=_('Document')
    )
    datetime_accessed = models.DateTimeField(
        auto_now=True, db_index=True, verbose_name=_('Accessed')
    )

    objects = RecentlyAccessedDocumentManager()

    class Meta:
        ordering = ('-datetime_accessed',)
        verbose_name = _('Recent document')
        verbose_name_plural = _('Recent documents')

    def __str__(self):
        return force_text(s=self.document)

    def natural_key(self):
        return (self.datetime_accessed, self.document.natural_key(), self.user.natural_key())
    natural_key.dependencies = ['documents.Document', settings.AUTH_USER_MODEL]



from rest_framework import serializers

from ..models.trashed_document_models import TrashedDocument

from .document_serializers import DocumentSerializer


class TrashedDocumentSerializer(DocumentSerializer):
    restore_url = serializers.HyperlinkedIdentityField(
        lookup_url_kwarg='document_id',
        view_name='rest_api:trasheddocument-restore'
    )

    class Meta:
        extra_kwargs = {
            'url': {
                'lookup_url_kwarg': 'document_id',
                'view_name': 'rest_api:trasheddocument-detail'
            }
        }
        fields = sorted(
            DocumentSerializer.Meta.fields + (
                'restore_url', 'trashed_date_time'
            )
        )
        model = TrashedDocument



from mayan.apps.testing.tests.base import GenericViewTestCase

from ..permissions import permission_error_log_view

from .literals import TEST_ERROR_LOG_ENTRY_RESULT
from .mixins import LoggingTextMixin, LoggingViewTestMixin


class LoggingViewTestCase(
    LoggingTextMixin, LoggingViewTestMixin, GenericViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_error_log_test_object()
        self._create_error_log_entry()

    def test_object_error_list_view_no_permission(self):
        response = self._request_object_error_log_list_view()
        self.assertNotContains(
            response=response, text=TEST_ERROR_LOG_ENTRY_RESULT,
            status_code=404
        )

    def test_object_error_list_view_with_access(self):
        self.grant_access(
            obj=self.test_object, permission=permission_error_log_view
        )

        response = self._request_object_error_log_list_view()
        self.assertContains(
            response=response, text=TEST_ERROR_LOG_ENTRY_RESULT,
            status_code=200
        )

    def test_object_error_list_clear_view_no_permission(self):
        response = self._request_object_error_log_list_clear_view()
        self.assertEqual(response.status_code, 404)

        self.assertNotEqual(self.test_object.error_log.count(), 0)

    def test_object_error_list_clear_view_with_access(self):
        self.grant_access(
            obj=self.test_object, permission=permission_error_log_view
        )

        response = self._request_object_error_log_list_clear_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(self.test_object.error_log.count(), 0)



from mayan.apps.common.tests.mixins import ObjectCopyTestMixin
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from .mixins.workflow_template_mixins import WorkflowTemplateTestMixin
from .mixins.workflow_template_state_mixins import WorkflowTemplateStateActionTestMixin


class WorkflowTemplateCopyTestCase(
    DocumentTestMixin, ObjectCopyTestMixin, WorkflowTemplateStateActionTestMixin,
    WorkflowTemplateTestMixin, BaseTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_workflow_template()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()
        self._create_test_workflow_template_state_action()
        self.test_workflow_template.document_types.add(
            self.test_document_type
        )
        self.test_object = self.test_workflow_template



from django.conf.urls import url

from .views import (
    CheckVersionView, DependencyGroupEntryListView,
    DependencyGroupEntryDetailView, DependencyGroupListView,
    DependencyLicensesView
)

urlpatterns = [
    url(
        regex=r'^check_version/$', name='check_version_view',
        view=CheckVersionView.as_view()
    ),
    url(
        regex=r'^groups/(?P<dependency_group_name>\w+)/(?P<dependency_group_entry_name>\w+)$',
        name='dependency_group_entry_detail',
        view=DependencyGroupEntryDetailView.as_view()
    ),
    url(
        regex=r'^groups/(?P<dependency_group_name>\w+)/$',
        name='dependency_group_entry_list',
        view=DependencyGroupEntryListView.as_view()
    ),
    url(
        regex=r'^groups/$', name='dependency_group_list',
        view=DependencyGroupListView.as_view()
    ),
    url(
        regex=r'^licenses/$', name='dependency_licenses_view',
        view=DependencyLicensesView.as_view()
    )
]



from ..permissions import (
    permission_document_file_view, permission_document_view
)

from .base import GenericDocumentViewTestCase
from .mixins.document_mixins import DocumentSearchTestMixin


class DocumentSearchTestCase(
    DocumentSearchTestMixin, GenericDocumentViewTestCase
):
    def test_document_file_page_search_no_permission(self):
        queryset = self._perform_document_file_page_search()
        self.assertFalse(
            self.test_document.file_latest.pages.first() in queryset
        )

    def test_document_file_page_search_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_file_view
        )
        queryset = self._perform_document_file_page_search()
        self.assertTrue(
            self.test_document.file_latest.pages.first() in queryset
        )

    def test_document_search_no_permission(self):
        queryset = self._perform_document_search()
        self.assertFalse(self.test_document in queryset)

    def test_document_search_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )
        queryset = self._perform_document_search()
        self.assertTrue(self.test_document in queryset)



import logging
import os

from django import forms
from django.utils.translation import ugettext_lazy as _

from mayan.apps.views.forms import DetailForm

from ..models.document_models import Document
from ..settings import setting_language
from ..utils import get_language, get_language_choices

__all__ = ('DocumentForm', 'DocumentPropertiesForm',)
logger = logging.getLogger(name=__name__)


class DocumentForm(forms.ModelForm):
    """
    Base form for the minimal document properties. Meant to be subclassed.
    """
    class Meta:
        fields = ('label', 'description', 'language')
        model = Document

    def __init__(self, *args, **kwargs):
        document_type = kwargs.pop('document_type', None)

        super().__init__(*args, **kwargs)

        # Is a document (documents app edit) and has been saved (sources
        # app upload)?
        if self.instance and self.instance.pk:
            document_type = self.instance.document_type
        else:
            self.initial.update({'language': setting_language.value})

        filenames_queryset = document_type.filenames.filter(enabled=True)

        if filenames_queryset:
            self.fields[
                'document_type_available_filenames'
            ] = forms.ModelChoiceField(
                queryset=filenames_queryset,
                required=False,
                label=_('Quick document rename'),
                widget=forms.Select(
                    attrs={
                        'class': 'select2'
                    }
                )
            )
            self.fields['preserve_extension'] = forms.BooleanField(
                label=_('Preserve extension'), required=False,
                help_text=_(
                    'Takes the file extension and moves it to the end of the '
                    'filename allowing operating systems that rely on file '
                    'extensions to open document correctly.'
                )
            )

        self.fields['language'].widget = forms.Select(
            choices=get_language_choices(), attrs={
                'class': 'select2'
            }
        )

    def clean(self):
        self.cleaned_data['label'] = self.get_final_label(
            # Fallback to the instance label if there is no label key or
            # there is a label key and is an empty string
            filename=self.cleaned_data.get('label') or self.instance.label
        )

        return self.cleaned_data

    def get_final_label(self, filename):
        if 'document_type_available_filenames' in self.cleaned_data:
            if self.cleaned_data['document_type_available_filenames']:
                if self.cleaned_data['preserve_extension']:
                    filename, extension = os.path.splitext(filename)

                    filename = '{}{}'.format(
                        self.cleaned_data[
                            'document_type_available_filenames'
                        ].filename, extension
                    )
                else:
                    filename = self.cleaned_data[
                        'document_type_available_filenames'
                    ].filename

        return filename


class DocumentPropertiesForm(DetailForm):
    """
    Detail class form to display a document file based properties
    """
    def __init__(self, *args, **kwargs):
        document = kwargs['instance']

        extra_fields = [
            {
                'label': _('Date created'),
                'field': 'datetime_created',
                'widget': forms.widgets.DateTimeInput
            },
            {'label': _('UUID'), 'field': 'uuid'},
            {
                'label': _('Language'),
                'field': lambda x: get_language(language_code=document.language)
            },
        ]

        kwargs['extra_fields'] = extra_fields
        super().__init__(*args, **kwargs)

    class Meta:
        fields = ('document_type', 'description')
        model = Document



from django.urls import resolve
from django.urls.exceptions import Resolver404

from mayan.apps.organizations.settings import setting_organization_installation_url

from .literals import API_VERSION


class Endpoint:
    def __init__(self, label, viewname=None, kwargs=None):
        self.label = label
        self.kwargs = kwargs

        if viewname:
            self.viewname = viewname
        else:
            installation_base_url = setting_organization_installation_url.value
            if installation_base_url:
                installation_base_url = '/{}'.format(installation_base_url)
            else:
                installation_base_url = ''

            self.url = '{}/api/v{}/{}/'.format(
                installation_base_url, API_VERSION, self.label
            )

            try:
                self.viewname = resolve(path=self.url).view_name
            except Resolver404:
                self.viewname = None



import glob
import importlib
import logging
import os
import random
import time

from furl import furl
from selenium.webdriver.firefox.options import Options
from selenium.webdriver.firefox.webdriver import WebDriver

from django.apps import apps
from django.conf import settings
from django.conf.urls import url
from django.contrib.contenttypes.models import ContentType
from django.db import connection, connections, models
from django.db.models.signals import post_save, pre_save
from django.http import HttpResponse
from django.template import Context, Template
from django.test.utils import ContextList
from django.urls import clear_url_caches, reverse
from django.utils.encoding import DjangoUnicodeDecodeError, force_text

from stronghold.decorators import public

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.permissions.tests.mixins import PermissionTestMixin
from mayan.apps.storage.settings import setting_temporary_directory
from mayan.apps.views.compat import FileResponse

from ..literals import (
    TEST_SERVER_HOST, TEST_SERVER_SCHEME, TEST_VIEW_NAME, TEST_VIEW_URL
)


if getattr(settings, 'COMMON_TEST_FILE_HANDLES', False):
    import psutil


class ClientMethodsTestCaseMixin:
    def _build_verb_kwargs(self, viewname=None, path=None, *args, **kwargs):
        data = kwargs.pop('data', None) or {}
        follow = kwargs.pop('follow', False)
        query = kwargs.pop('query', None) or {}
        headers = kwargs.pop('headers', None) or {}

        if viewname:
            path = reverse(viewname=viewname, *args, **kwargs)

        path = furl(url=path)
        path.args.update(query)

        result = {'follow': follow, 'data': data, 'path': path.tostr()}
        result.update(headers)
        return result

    def delete(self, viewname=None, path=None, *args, **kwargs):
        return self.client.delete(
            **self._build_verb_kwargs(
                path=path, viewname=viewname, *args, **kwargs
            )
        )

    def generic(self, method, viewname=None, path=None, *args, **kwargs):
        return self.client.generic(
            method=method, **self._build_verb_kwargs(
                path=path, viewname=viewname, *args, **kwargs
            )
        )

    def get(self, viewname=None, path=None, *args, **kwargs):
        return self.client.get(
            **self._build_verb_kwargs(
                path=path, viewname=viewname, *args, **kwargs
            )
        )

    def patch(self, viewname=None, path=None, *args, **kwargs):
        return self.client.patch(
            **self._build_verb_kwargs(
                path=path, viewname=viewname, *args, **kwargs
            )
        )

    def post(self, viewname=None, path=None, *args, **kwargs):
        return self.client.post(
            **self._build_verb_kwargs(
                path=path, viewname=viewname, *args, **kwargs
            )
        )

    def put(self, viewname=None, path=None, *args, **kwargs):
        return self.client.put(
            **self._build_verb_kwargs(
                path=path, viewname=viewname, *args, **kwargs
            )
        )


class ConnectionsCheckTestCaseMixin:
    _open_connections_check_enable = True

    def _get_open_connections_count(self):
        return len(connections.all())

    def setUp(self):
        super().setUp()
        self._connections_count = self._get_open_connections_count()

    def tearDown(self):
        if self._open_connections_check_enable:

            self.assertEqual(
                self._connections_count, self._get_open_connections_count(),
                msg='Database connection leak. The number of database '
                'connections at the start and at the end of the test are not '
                'the same.'
            )

        super().tearDown()


class ContentTypeCheckTestCaseMixin:
    expected_content_types = ('text/html', 'text/html; charset=utf-8')

    def _pre_setup(self):
        super()._pre_setup()
        test_instance = self

        class CustomClient(self.client_class):
            def request(self, *args, **kwargs):
                response = super().request(*args, **kwargs)

                content_type = response._headers.get('content-type', [None, ''])[1]
                if test_instance.expected_content_types:
                    test_instance.assertTrue(
                        content_type in test_instance.expected_content_types,
                        msg='Unexpected response content type: {}, expected: {}.'.format(
                            content_type, ' or '.join(test_instance.expected_content_types)
                        )
                    )

                return response

        self.client = CustomClient()


class ContentTypeTestCaseMixin:
    def _inject_test_object_content_type(self):
        self.test_object_content_type = ContentType.objects.get_for_model(
            model=self.test_object
        )

        self.test_object_view_kwargs = {
            'app_label': self.test_object_content_type.app_label,
            'model_name': self.test_object_content_type.model,
            'object_id': self.test_object.pk
        }


class DelayTestCaseMixin:
    def _test_delay(self, seconds=0.1):
        time.sleep(seconds)


class DownloadTestCaseMixin:
    def assert_download_response(
        self, response, content=None, filename=None, is_attachment=None,
        mime_type=None
    ):
        self.assertTrue(isinstance(response, FileResponse))

        if filename:
            self.assertEqual(
                response[
                    'Content-Disposition'
                ].split('filename="')[1].split('"')[0], filename
            )

        if content:
            response_content = b''.join(list(response))

            try:
                response_content = force_text(s=response_content)
            except DjangoUnicodeDecodeError:
                """Leave as bytes"""

            self.assertEqual(response_content, content)

        if is_attachment is not None:
            self.assertEqual(response['Content-Disposition'], 'attachment')

        if mime_type:
            self.assertTrue(response['Content-Type'].startswith(mime_type))


class EnvironmentTestCaseMixin:
    def setUp(self):
        super().setUp()
        self._test_environment_variables = []

    def tearDown(self):
        for name in self._test_environment_variables:
            os.environ.pop(name)

        super().tearDown()

    def _set_environment_variable(self, name, value):
        self._test_environment_variables.append(name)
        os.environ[name] = value


class ModelTestCaseMixin:
    def _model_instance_to_dictionary(self, instance):
        return instance._meta.model._default_manager.filter(
            pk=instance.pk
        ).values()[0]


class OpenFileCheckTestCaseMixin:
    def _get_descriptor_count(self):
        process = psutil.Process()
        return process.num_fds()

    def _get_open_files(self):
        process = psutil.Process()
        return process.open_files()

    def setUp(self):
        super().setUp()
        if getattr(settings, 'COMMON_TEST_FILE_HANDLES', False):
            self._open_files = self._get_open_files()

    def tearDown(self):
        if getattr(settings, 'COMMON_TEST_FILE_HANDLES', False) and not getattr(self, '_skip_file_descriptor_test', False):
            for new_open_file in self._get_open_files():
                self.assertFalse(
                    new_open_file not in self._open_files,
                    msg='File descriptor leak. The number of file descriptors '
                    'at the start and at the end of the test are not the same.'
                )

            self._skip_file_descriptor_test = False

        super().tearDown()


class RandomPrimaryKeyModelMonkeyPatchMixin:
    random_primary_key_random_floor = 100
    random_primary_key_random_ceiling = 10000
    random_primary_key_maximum_attempts = 100
    random_primary_key_enable = True

    @staticmethod
    def get_unique_primary_key(model):
        manager = model._meta.default_manager

        attempts = 0
        while True:
            primary_key = random.randint(
                RandomPrimaryKeyModelMonkeyPatchMixin.random_primary_key_random_floor,
                RandomPrimaryKeyModelMonkeyPatchMixin.random_primary_key_random_ceiling
            )

            if not manager.filter(pk=primary_key).exists():
                break

            attempts = attempts + 1

            if attempts > RandomPrimaryKeyModelMonkeyPatchMixin.random_primary_key_maximum_attempts:
                raise Exception(
                    'Maximum number of retries for an unique random primary '
                    'key reached.'
                )

        return primary_key

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        random.seed()

    def setUp(self):
        if self.random_primary_key_enable:
            self.method_save_original = models.Model.save

            def method_save_new(instance, *args, **kwargs):
                if instance.pk:
                    return self.method_save_original(instance, *args, **kwargs)
                else:
                    # Set meta.auto_created to True to have the original save_base
                    # not send the pre_save signal which would normally send
                    # the instance without a primary key. Since we assign a random
                    # primary key any pre_save signal handler that relies on an
                    # empty primary key will fail.
                    # The meta.auto_created and manual pre_save sending emulates
                    # the original behavior. Since meta.auto_created also disables
                    # the post_save signal we must also send it ourselves.
                    # This hack work with Django 1.11 .save_base() but can break
                    # in future versions if that method is updated.
                    pre_save.send(
                        sender=instance.__class__, instance=instance, raw=False,
                        update_fields=None,
                    )
                    instance._meta.auto_created = True
                    instance.pk = RandomPrimaryKeyModelMonkeyPatchMixin.get_unique_primary_key(
                        model=instance._meta.model
                    )
                    instance.id = instance.pk

                    kwargs['force_insert'] = True

                    result = instance.save_base(*args, **kwargs)
                    instance._meta.auto_created = False

                    post_save.send(
                        sender=instance.__class__, instance=instance, created=True,
                        update_fields=None, raw=False
                    )

                    return result

            setattr(models.Model, 'save', method_save_new)
        super().setUp()

    def tearDown(self):
        if self.random_primary_key_enable:
            models.Model.save = self.method_save_original
        super().tearDown()


class SeleniumTestMixin:
    SKIP_VARIABLE_NAME = 'TESTS_SELENIUM_SKIP'

    @staticmethod
    def _get_skip_variable_value():
        return os.environ.get(
            SeleniumTestMixin._get_skip_variable_environment_name(),
            getattr(settings, SeleniumTestMixin.SKIP_VARIABLE_NAME, False)
        )

    @staticmethod
    def _get_skip_variable_environment_name():
        return 'MAYAN_{}'.format(SeleniumTestMixin.SKIP_VARIABLE_NAME)

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.webdriver = None
        if not SeleniumTestMixin._get_skip_variable_value():
            options = Options()
            options.add_argument('--headless')
            cls.webdriver = WebDriver(
                firefox_options=options, log_path='/dev/null'
            )

    @classmethod
    def tearDownClass(cls):
        if cls.webdriver:
            cls.webdriver.quit()
        super().tearDownClass()

    def setUp(self):
        if SeleniumTestMixin._get_skip_variable_value():
            self.skipTest(reason='Skipping selenium test')
        super().setUp()

    def _open_url(self, fragment=None, path=None, viewname=None):
        url = '{}{}{}'.format(
            self.live_server_url, path or reverse(viewname=viewname),
            fragment or ''
        )

        self.webdriver.get(url=url)


class SilenceLoggerTestCaseMixin:
    """
    Changes the log level of a specific logger for the duration of a test.
    The default level for silenced loggers is CRITICAL.
    Example: self._silence_logger(name='mayan.apps.converter.managers')
    """
    test_case_silenced_logger = None
    test_case_silenced_logger_new_level = logging.CRITICAL

    def tearDown(self):
        if self.test_case_silenced_logger:
            self.test_case_silenced_logger.setLevel(
                level=self.test_case_silenced_logger_level
            )

        super().tearDown()

    def _silence_logger(self, name):
        self.test_case_silenced_logger = logging.getLogger(name=name)
        self.test_case_silenced_logger_level = self.test_case_silenced_logger.level
        self.test_case_silenced_logger.setLevel(
            level=self.test_case_silenced_logger_new_level
        )


class TempfileCheckTestCasekMixin:
    # Ignore the jvmstat instrumentation and GitLab's CI .config files
    # Ignore LibreOffice fontconfig cache dir
    ignore_globs = ('hsperfdata_*', '.config', '.cache')

    def _get_temporary_entries(self):
        ignored_result = []

        # Expand globs by joining the temporary directory and then flattening
        # the list of lists into a single list
        for item in self.ignore_globs:
            ignored_result.extend(
                glob.glob(
                    os.path.join(setting_temporary_directory.value, item)
                )
            )

        # Remove the path and leave only the expanded filename
        ignored_result = map(lambda x: os.path.split(x)[-1], ignored_result)

        return set(
            os.listdir(setting_temporary_directory.value)
        ) - set(ignored_result)

    def setUp(self):
        super().setUp()
        if getattr(settings, 'COMMON_TEST_TEMP_FILES', False):
            self._temporary_items = self._get_temporary_entries()

    def tearDown(self):
        if getattr(settings, 'COMMON_TEST_TEMP_FILES', False):
            final_temporary_items = self._get_temporary_entries()
            self.assertEqual(
                self._temporary_items, final_temporary_items,
                msg='Orphan temporary file. The number of temporary files and/or '
                'directories at the start and at the end of the test are not the '
                'same. Orphan entries: {}'.format(
                    ','.join(final_temporary_items - self._temporary_items)
                )
            )
        super().tearDown()


class TestModelTestCaseMixin(ContentTypeTestCaseMixin, PermissionTestMixin):
    auto_create_test_object = False
    auto_create_test_object_fields = None
    auto_create_test_object_instance_kwargs = None
    auto_create_test_object_permission = False

    @classmethod
    def setUpClass(cls):
        if connection.vendor == 'sqlite':
            connection.disable_constraint_checking()

        super().setUpClass()

    def setUp(self):
        self._test_models = []
        self.test_objects = []

        super().setUp()

        if self.auto_create_test_object:
            self._create_test_object(
                fields=self.auto_create_test_object_fields,
                create_test_permission=self.auto_create_test_object_permission,
                instance_kwargs=self.auto_create_test_object_instance_kwargs
            )

    def tearDown(self):
        # Delete the test models' content type entries and deregister the
        # permissions, this avoids their Content Type from being looked up
        # in subsequent tests where they don't exists due to the database
        # transaction rollback.
        for model in self._test_models:
            content_type = ContentType.objects.get_for_model(model=model)
            if content_type.pk:
                content_type.delete()
            ModelPermission.deregister(model=model)

        super().tearDown()

    def _create_test_model(
        self, base_class=models.Model, fields=None, model_name=None,
        options=None
    ):
        test_model_count = len(self._test_models)
        self._test_model_name = model_name or '{}_{}'.format(
            'TestModel', test_model_count
        )

        self.options = options
        # Obtain the app_config and app_label from the test's module path
        self.app_config = apps.get_containing_app_config(
            object_name=self.__class__.__module__
        )

        if connection.vendor == 'mysql':
            self.skipTest(
                reason='MySQL doesn\'t support schema changes inside an '
                'atomic block.'
            )

        attrs = {
            '__module__': self.__class__.__module__,
            'save': self._get_test_model_save_method(),
            'Meta': self._get_test_model_meta(),
        }

        if fields:
            attrs.update(fields)

        # Clear previous model registration before re-registering it again to
        # avoid conflict with test models with the same name, in the same app
        # but from another test module.
        apps.all_models[self.app_config.label].pop(
            self._test_model_name.lower(), None
        )

        model = type(
            self._test_model_name, (base_class,), attrs
        )

        if not model._meta.proxy:
            with connection.schema_editor() as schema_editor:
                schema_editor.create_model(model=model)

        self._test_models.append(model)
        ContentType.objects.clear_cache()

        return model

    def _create_test_object(
        self, fields=None, model_name=None, create_test_permission=False,
        instance_kwargs=None
    ):
        instance_kwargs = instance_kwargs or {}

        self.TestModel = self._create_test_model(
            fields=fields, model_name=model_name
        )
        self.test_object = self.TestModel.objects.create(**instance_kwargs)
        self._inject_test_object_content_type()

        if create_test_permission:
            self._create_test_permission()

            ModelPermission.register(
                model=self.TestModel, permissions=(
                    self.test_permission,
                )
            )

        self.test_objects.append(self.test_object)

    def _get_test_model_meta(self):
        self._test_db_table = '{}_{}'.format(
            self.app_config.label, self._test_model_name.lower()
        )

        class Meta:
            app_label = self.app_config.label
            db_table = self._test_db_table
            verbose_name = self._test_model_name

        if self.options:
            for key, value in self.options.items():
                setattr(Meta, key, value)

        return Meta

    def _get_test_model_save_method(self):
        def save(instance, *args, **kwargs):
            # Custom .save() method to use random primary key values.
            if instance.pk:
                return models.Model.self(instance, *args, **kwargs)
            else:
                instance.pk = RandomPrimaryKeyModelMonkeyPatchMixin.get_unique_primary_key(
                    model=instance._meta.model
                )
                instance.id = instance.pk

                kwargs['force_insert'] = True
                return instance.save_base(*args, **kwargs)
        return save


class TestServerTestCaseMixin:
    def setUp(self):
        super().setUp()
        self.testserver_prefix = self.get_testserver_prefix()
        self.testserver_url = self.get_testserver_url()
        self.test_view_request = None

    def _test_view_factory(self, test_object=None):
        def test_view(request):
            self.test_view_request = request
            return HttpResponse()

        return test_view

    def get_testserver_prefix(self):
        return furl(
            scheme=TEST_SERVER_SCHEME, host=TEST_SERVER_HOST,
        ).tostr()

    def get_testserver_url(self):
        return furl(
            scheme=TEST_SERVER_SCHEME, host=TEST_SERVER_HOST,
            path=self.test_view_url
        ).tostr()


class TestViewTestCaseMixin:
    auto_add_test_view = False
    has_test_view = False
    test_view_is_public = False
    test_view_object = None
    test_view_name = TEST_VIEW_NAME
    test_view_template = '{{ object }}'
    test_view_url = TEST_VIEW_URL

    def setUp(self):
        super().setUp()
        if self.auto_add_test_view:
            self.add_test_view(test_object=self.test_view_object)

    def tearDown(self):
        urlconf = importlib.import_module(settings.ROOT_URLCONF)

        self.client.logout()
        if self.has_test_view:
            urlconf.urlpatterns.pop(0)
        super().tearDown()

    def _get_context_from_test_response(self, response):
        if isinstance(response.context, ContextList):
            # template widget rendering causes test client response to be
            # ContextList rather than RequestContext. Typecast to dictionary
            # before updating.
            result = dict(response.context).copy()
            result.update({'request': response.wsgi_request})
            context = Context(result)
        else:
            result = response.context or {}
            result.update({'request': response.wsgi_request})
            context = Context(result)

        context.request = response.wsgi_request
        return context

    def _test_view_factory(self, test_object=None):
        def test_view(request):
            template = Template(template_string=self.test_view_template)
            context = Context(
                dict_={'object': test_object, 'resolved_object': test_object}
            )
            return HttpResponse(template.render(context=context))

        if self.test_view_is_public:
            return public(function=test_view)
        else:
            return test_view

    def add_test_view(self, test_object=None):
        urlconf = importlib.import_module(settings.ROOT_URLCONF)

        urlconf.urlpatterns.insert(
            0, url(
                regex=self.test_view_url, view=self._test_view_factory(
                    test_object=test_object
                ), name=self.test_view_name
            )
        )
        clear_url_caches()
        self.has_test_view = True

    def get_test_view(self):
        response = self.get(viewname=self.test_view_name)
        return self._get_context_from_test_response(response=response)



class GPGException(Exception):
    """
    Base app exception.
    """


class DecryptionError(GPGException):
    """
    Raised when an error occurs while trying to decrypt and encrypted file.
    """


class KeyFetchingError(GPGException):
    """
    Unable to receive key or key not found.
    """


class KeyDoesNotExist(GPGException):
    """
    Raised when attempting to decrypt or verify a file, and then key used to
    encrypt or sign the file is not found.
    """


class NeedPassphrase(GPGException):
    """
    A passphrase is needed but none was provided.
    """


class PassphraseError(GPGException):
    """
    Passphrase provided is incorrect.
    """


class VerificationError(GPGException):
    """
    Raised when a file is not signed.
    """



from mayan.apps.events.classes import EventManagerMethodAfter
from mayan.apps.events.decorators import method_event

from ..events import event_trashed_document_restored

from ..managers import TrashCanManager

from .document_models import Document

__all__ = ('TrashedDocument',)


class TrashedDocument(Document):
    objects = TrashCanManager()

    class Meta:
        proxy = True

    @method_event(
        event_manager_class=EventManagerMethodAfter,
        event=event_trashed_document_restored,
        target='self',
    )
    def restore(self):
        self.in_trash = False
        # Skip the edite event at .save().
        self._event_ignore = True
        self.save()



import logging

from django.utils import six
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.class_mixins import AppsModuleLoaderMixin

logger = logging.getLogger(name=__name__)


__all__ = ('MailerBackend',)


class MailerBackendMetaclass(type):
    _registry = {}

    def __new__(mcs, name, bases, attrs):
        new_class = super().__new__(
            mcs, name, bases, attrs
        )
        if not new_class.__module__ == 'mayan.apps.mailer.classes':
            mcs._registry[
                '{}.{}'.format(new_class.__module__, name)
            ] = new_class

        return new_class


class MailerBackendBase(AppsModuleLoaderMixin):
    """
    Base class for the mailing backends. This class is mainly a wrapper
    for other Django backends that adds a few metadata to specify the
    fields it needs to be instanciated at runtime.

    The fields attribute is a list of dictionaries with the format:
    {
        'name': ''  # Field internal name
        'label': ''  # Label to show to users
        'class': ''  # Field class to use. Field classes are Python dot
                       paths to Django's form fields.
        'initial': ''  # Field initial value
        'default': ''  # Default value.
    }

    """
    class_path = ''  # Dot path to the actual class that will handle the mail
    fields = {}

    @classmethod
    def get_class_fields(cls):
        backend_field_list = getattr(cls, 'fields', {}).keys()
        return getattr(cls, 'class_fields', backend_field_list)


class MailerBackend(
    six.with_metaclass(MailerBackendMetaclass, MailerBackendBase)
):
    _loader_module_name = 'mailers'

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    @classmethod
    def get_all(cls):
        return cls._registry

    @classmethod
    def get_choices(cls):
        return sorted(
            [
                (
                    key, backend.label
                ) for key, backend in cls.get_all().items()
            ], key=lambda x: x[1]
        )


class NullBackend(MailerBackend):
    label = _('Null backend')



from django.template import Library

from ..classes import Dashboard

register = Library()


@register.simple_tag(takes_context=True)
def dashboards_render_dashboard(context, name):
    return Dashboard.get(name=name).render(request=context.request)



from mayan.apps.dependencies.classes import PythonDependency

PythonDependency(
    module=__name__, name='extract-msg', version_string='==0.23.3'
)
PythonDependency(
    module=__name__, name='pycryptodome', version_string='==3.9.7'
)



from django.utils.encoding import force_text

from mayan.apps.documents.permissions import (
    permission_document_file_view, permission_document_version_view,
    permission_document_view
)
from mayan.apps.documents.search import (
    document_file_page_search, document_file_search, document_search,
    document_version_page_search, document_version_search
)
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.dynamic_search.tests.mixins import SearchViewTestMixin

from ..permissions import permission_document_metadata_view

from .mixins import DocumentMetadataMixin, MetadataTypeTestMixin


class DocumentSearchResultWidgetViewTestCase(
    DocumentMetadataMixin, MetadataTypeTestMixin, SearchViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_metadata_type(add_test_document_type=True)
        self._upload_test_document()
        self._create_test_document_metadata()
        self._test_object_permission = permission_document_view
        self._test_object_text = self.test_document.label
        self._test_search_model = document_search
        self._test_search_term_data = {'uuid': self.test_document.uuid}

    def test_document_metadata_widget_no_permission(self):
        response = self._request_search_results_view(
            data=self._test_search_term_data, kwargs={
                'search_model_name': self._test_search_model.get_full_name()
            }
        )
        self.assertNotContains(
            response=response, text=self._test_object_text,
            status_code=200
        )
        self.assertNotContains(
            response=response, text=self.test_metadata_type.label,
            status_code=200
        )

    def test_document_metadata_widget_with_metadata_type_access(self):
        self.grant_access(
            obj=self.test_metadata_type,
            permission=permission_document_metadata_view
        )

        response = self._request_search_results_view(
            data=self._test_search_term_data, kwargs={
                'search_model_name': self._test_search_model.get_full_name()
            }
        )
        self.assertNotContains(
            response=response, text=self._test_object_text, status_code=200
        )
        self.assertNotContains(
            response=response, text=self.test_metadata_type.label,
            status_code=200
        )

    def test_document_metadata_widget_with_document_view_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=self._test_object_permission
        )

        response = self._request_search_results_view(
            data=self._test_search_term_data, kwargs={
                'search_model_name': self._test_search_model.get_full_name()
            }
        )
        self.assertContains(
            response=response, text=self._test_object_text, status_code=200
        )
        self.assertNotContains(
            response=response, text=self.test_metadata_type.label,
            status_code=200
        )

    def test_document_metadata_widget_with_document_metadata_view_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=permission_document_metadata_view
        )

        response = self._request_search_results_view(
            data=self._test_search_term_data, kwargs={
                'search_model_name': self._test_search_model.get_full_name()
            }
        )
        self.assertNotContains(
            response=response, text=self._test_object_text, status_code=200
        )
        self.assertNotContains(
            response=response, text=self.test_metadata_type.label,
            status_code=200
        )

    def test_document_metadata_widget_with_all_document_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=self._test_object_permission
        )
        self.grant_access(
            obj=self.test_document,
            permission=permission_document_metadata_view
        )

        response = self._request_search_results_view(
            data=self._test_search_term_data, kwargs={
                'search_model_name': self._test_search_model.get_full_name()
            }
        )
        self.assertContains(
            response=response, text=self._test_object_text, status_code=200
        )
        self.assertNotContains(
            response=response, text=self.test_metadata_type.label,
            status_code=200
        )

    def test_document_metadata_widget_with_metadata_view_and_document_view_access(self):
        self.grant_access(
            obj=self.test_document, permission=self._test_object_permission
        )
        self.grant_access(
            obj=self.test_metadata_type,
            permission=permission_document_metadata_view
        )

        response = self._request_search_results_view(
            data=self._test_search_term_data, kwargs={
                'search_model_name': self._test_search_model.get_full_name()
            }
        )
        self.assertContains(
            response=response, text=self._test_object_text, status_code=200
        )
        self.assertNotContains(
            response=response, text=self.test_metadata_type.label,
            status_code=200
        )

    def test_document_metadata_widget_with_full_access(self):
        self.grant_access(
            obj=self.test_document,
            permission=self._test_object_permission
        )
        self.grant_access(
            obj=self.test_document,
            permission=permission_document_metadata_view
        )
        self.grant_access(
            obj=self.test_metadata_type,
            permission=permission_document_metadata_view
        )

        response = self._request_search_results_view(
            data=self._test_search_term_data, kwargs={
                'search_model_name': self._test_search_model.get_full_name()
            }
        )
        self.assertContains(
            response=response, text=self._test_object_text, status_code=200
        )
        self.assertContains(
            response=response, text=self.test_metadata_type.label,
            status_code=200
        )


class DocumentFileSearchResultWidgetViewTestCase(
    DocumentSearchResultWidgetViewTestCase
):
    def setUp(self):
        super().setUp()
        self._test_object_text = self.test_document_file.filename
        self._test_object_permission = permission_document_file_view
        self._test_search_model = document_file_search
        self._test_search_term_data = {
            'document__uuid': self.test_document.uuid
        }


class DocumentFilePageSearchResultWidgetViewTestCase(
    DocumentSearchResultWidgetViewTestCase
):
    def setUp(self):
        super().setUp()
        self._test_object_text = force_text(s=self.test_document_file.pages.first())
        self._test_object_permission = permission_document_file_view
        self._test_search_model = document_file_page_search
        self._test_search_term_data = {
            'document_file__document__uuid': self.test_document.uuid
        }


class DocumentVersionSearchResultWidgetViewTestCase(
    DocumentSearchResultWidgetViewTestCase
):
    def setUp(self):
        super().setUp()
        self._test_object_text = force_text(s=self.test_document_version)
        self._test_object_permission = permission_document_version_view
        self._test_search_model = document_version_search
        self._test_search_term_data = {
            'document__uuid': self.test_document.uuid
        }


class DocumentVersionPageSearchResultWidgetViewTestCase(
    DocumentSearchResultWidgetViewTestCase
):
    def setUp(self):
        super().setUp()
        self._test_object_text = force_text(s=self.test_document_version.pages.first())
        self._test_object_permission = permission_document_version_view
        self._test_search_model = document_version_page_search
        self._test_search_term_data = {
            'document_version__document__uuid': self.test_document.uuid
        }
